/*			
**			         HTNFS.h
**			================================
**
**  module written by
**    Ashish Singhal <asinghal@cs.stanford.edu, ashish.singhal@eng.sun.com>
**    Sun Microsystems
**    Mountain View.
*/

#ifndef HTNFS_H
#define HTNFS_H

#include "HTUtils.h"
#include "HTAnchor.h"
#include "HTStream.h"
#include <rpc/rpc.h>


extern int HTNFSLoad PARAMS
((
  char *          name,
  HTParentAnchor *      anchor,
  HTFormat              format_out,
  HTStream*             sink
));


/*
Return Host Name
 */
extern CONST char * HTHostName NOPARAMS;


/* some constants */
#define NFS_STARTOVER         -2001
#define NFS_MOUNTPOSSIBLE     -2002

#define IPPORT_NFS            2049
#define NFSCLIENTCACHESIZE    10                    /* max no of cached 
                                                     * client              */
#define MAX_NFS3_PACKET       ( 32 * 1024 )         /* 32K                 */
#define MAX_NFS3_READ_COUNT   MAX_NFS3_PACKET
#define BUFSIZE               1024
#define MAX_NFS3_DIRPLUS_INFO (MAX_NFS3_PACKET / 8) /* 1/8th of max packet */
#define MAX_NFS2_READ_COUNT   ( 8 * 1024 )


/* nfs errors */
enum nfsError {
  NO_ERROR      = 0,
  NFS_ERR_RPC   = 1,
  NFS_ERR_STALE = 2,
  NFS_ERR_MNT   = 3
};


/* struct for sorting the directory entries */
struct _readDir {
  char          *filename;       /* filename                   */
  u_longlong_t  size;            /* size of the file           */
  int           type;            /* type of the file           */
  struct _readDir  *next;
};
typedef struct _readDir readDir;


/* struct for caching pathnames for a host  */
struct _prefix {
  char      *pathname;           /* pathname                   */
  void      *fh;                 /* file handle (both v3 & v2  */
  struct _prefix  *next;
};
typedef struct _prefix prefix;


/* struct for caching the client handles    */
struct _nfsCache {
  CLIENT        *clnt;           /* client handle              */
  int            fd;             /* socket descriptor for client
                                  * connection                 */
  struct t_bind *tbind;          /* bind info                  */
  char          *hostname;       /* hostname                   */
  int            Version;        /* NFS version of the client  */
  BOOL           PublicFH;       /* flag for checking if the 
                                  * host supports public fh    */
  CLIENT        *MountClnt;      /* handle for mount proto     */
  prefix        *prefixtable;    /* prefix table               */
  struct _nfsCache  *next;
  struct _nfsCache  *prev;
};
typedef struct _nfsCache nfsCache;


#endif

