/*			
** 		      NFS mount/unmodified server handling
**	              ====================================
**
**  module written by
**    Ashish Singhal <asinghal@cs.stanford.edu, ashish.singhal@eng.sun.com>
**    Sun Microsystems
**    Mountain View.
*/

#include "nfs_prot3.h"
#include "nfs_prot2.h"
#include "HTNFS.h"
#include "mount.h"
#include <rpc/rpc.h>
#include <stropts.h>
#include <sys/types.h>
#include "HTParse.h"
#include "HTUtils.h"
#include "tcp.h"
#include "HTTCP.h"
#include "HTAnchor.h"
#include "HTFile.h"
#include "HTSort.h"
#include "HText.h"
#ifdef __STDC__
#include <stdlib.h>
#endif

extern int        errCode;
extern nfsCache * nfsCachePtr;
extern FILE     * debugFile;
extern int        FirstTime;
extern enum nfsError nfs_error;


BOOL   nfs3Mount();
BOOL   nfs2Mount();
extern char * nfs3ReadLink();
extern char * nfs2ReadLink();
extern int    NFSDEBUGFLAG ;


/*      LookUp thru the directories step by step
**	----------------------------------------
**
** On entry,
**	filename	address of a file/dir
**      dir_fh          dir handle rel to which to parse filename
** On exit,
**	result_3        result of lookup of the file/dir
**                      NULL if error
*/
PRIVATE LOOKUP3res* nfs3singleLookUp
ARGS4 (
  char *,          hostname,
  char *,	   filename,
  nfs_fh3 *,       dir_fh,
  CLIENT *,        clnt
)
{
    LOOKUP3res  *result_3;
    LOOKUP3args  nfsproc3_lookup_3_arg;
    char        *singlePath;
    
    if ( filename[0] == '/' ) 
      memcpy( (void*) filename, (void*) &filename[1], strlen(filename) );
    if ( filename[strlen(filename)-1] == '/' )
      filename[strlen(filename)-1] = '\0';

    memcpy((void*) &nfsproc3_lookup_3_arg.what.dir,
	   (void*) dir_fh,
	   sizeof(nfs_fh3));
    
    /*  get the file handle  */
    do {
      /*  check for interruption  */
      if(HTCheckActiveIcon(0)) {
	HTProgress("Data transfer interrupted.");
	errCode = HT_INTERRUPTED;
	return NULL;
      }

      /*   get a single name */
      singlePath = strchr(filename, '/');

      if(singlePath != NULL)
	  *singlePath = '\0';
      if ( (nfsproc3_lookup_3_arg.what.name = strdup(filename)) == NULL) {
	  errCode = errno;
	  return NULL;
      }
      
      result_3 = nfsproc3_lookup_3(&nfsproc3_lookup_3_arg, clnt);
      if (result_3 == (LOOKUP3res *) NULL) {
	  struct rpc_err rpcErr;

	  free(nfsproc3_lookup_3_arg.what.name);
	  clnt_geterr(clnt, &rpcErr);
	  
	  if(NFSDEBUGFLAG) clnt_perror(clnt, "call failed");
	  /* check for expired connection   */
	  if(FirstTime &&
	     ((rpcErr.re_status == RPC_CANTRECV) )) {
	    nfs_error = NFS_ERR_RPC;
	    errCode = HT_NOT_LOADED;
	  }
	  else
	    errCode = HT_NOT_LOADED;
	  return NULL;
      }

      if (result_3->status != NFS3_OK) {
    	free(nfsproc3_lookup_3_arg.what.name);
	if(result_3->status == NFS3ERR_STALE) {
	    nfs_error = NFS_ERR_STALE;
	    errCode = HT_NOT_LOADED;
	  }
	else
	  errCode = HT_NOT_LOADED;
        return NULL;
      }

      /*  check if the 'fh' is a link  */
      if(result_3->LOOKUP3res_u.resok.obj_attributes.post_op_attr_u.attributes.type == NF3LNK) {

	 char *sym_lnk_data, *newName=NULL;

	 /*   read the link  */
	 if( (sym_lnk_data = nfs3ReadLink(hostname,
					  &result_3->LOOKUP3res_u.resok.object,
					  clnt))
	     == NULL) 
	   return NULL;

	 /*  if the symbolic link is wrt to the root dir, then the URL is 
	  *  redirected to that using the 'file' WWW protocol             */
	 if(sym_lnk_data[0] == '/') {
 	   extern char *redirecting_url;

	   StrAllocCopy(redirecting_url, "file://");
	   StrAllocCat (redirecting_url, hostname);
	   StrAllocCat (redirecting_url, sym_lnk_data);
	   errCode = HT_REDIRECTING;
	   return NULL;
	 }

	 if( (newName != NULL) && (*newName)) {
	   free(newName);
	   newName == NULL;
	 }

	 /*  create new filename */
         if( (newName = strdup(sym_lnk_data)) == NULL)
	     return NULL;
	 if(singlePath != NULL) {
	   StrAllocCat(newName, (singlePath +1) );
	   *singlePath = '/';
	 }
	 /*  update pointers  */
	 singlePath = filename = newName;
	 free(nfsproc3_lookup_3_arg.what.name);
      }
      else {
	/* update the 'dir' handle */
	memcpy((void*) &nfsproc3_lookup_3_arg.what.dir,
	       (void*) &result_3->LOOKUP3res_u.resok.object,
	       sizeof(nfs_fh3));
	if(singlePath != NULL) 
	  memcpy((void*)filename, (void*) singlePath, strlen(++singlePath)+1);
      }	/*  else  */
    } while (singlePath != NULL);
      
    free(nfsproc3_lookup_3_arg.what.name);
 
    return result_3;
}

/*      LookUp thru the directories step by step
**	----------------------------------------
**
** On entry,
**	filename	address of a file/dir
**      dir_fh          dir handle rel to which to parse filename
** On exit,
**	result_3        result of lookup of the file/dir
**                      NULL if error
*/

PRIVATE diropres* nfs2singleLookUp
ARGS4 (
  char *,          hostname,
  char *,	   filename,
  nfs_fh *,        dir_fh,
  CLIENT *,        clnt
)
{
    diropres    *result_3;
    diropargs    nfsproc_lookup_2_arg;
    char        *singlePath;
    
    if ( filename[0] == '/' ) 
      memcpy( (void*) filename, (void*) &filename[1], strlen(filename) );
    if ( filename[strlen(filename)-1] == '/' )
      filename[strlen(filename)-1] = '\0';

    memcpy((void*) &nfsproc_lookup_2_arg.dir,
	   (void*) dir_fh,
	   sizeof(nfs_fh));
    
    /*  get the file handle  */
    do {
      /*  check for interruption  */
      if(HTCheckActiveIcon(0)) {
	HTProgress("Data transfer interrupted.");
	errCode = HT_INTERRUPTED;
	return NULL;
      }

      /*   get a single name */
      singlePath = strchr(filename, '/');

      if(singlePath != NULL)
	  *singlePath = '\0';
      if ( (nfsproc_lookup_2_arg.name = strdup(filename)) == NULL) {
	  errCode = errno;
	  return NULL;
      }
      
      result_3 = nfsproc_lookup_2(&nfsproc_lookup_2_arg, clnt);
      if (result_3 == (diropres *) NULL) {
	  struct rpc_err rpcErr;

	  free(nfsproc_lookup_2_arg.name);
	  clnt_geterr(clnt, &rpcErr);

	  if(NFSDEBUGFLAG) clnt_perror(clnt, "call failed");
	  /* check for expired connection   */
	  if(FirstTime &&
	     ((rpcErr.re_status == RPC_CANTRECV) )) {
	    nfs_error = NFS_ERR_RPC;
	    errCode = HT_NOT_LOADED;
	  }
	  else
	    errCode = HT_NOT_LOADED;
        return NULL;
      }

      if (result_3->status != NFS3_OK) {
    	free(nfsproc_lookup_2_arg.name);
	if(result_3->status == NFSERR_STALE) {
	    nfs_error = NFS_ERR_STALE;
	    errCode = HT_NOT_LOADED;
	}
	else
	  errCode = HT_NOT_LOADED;
        return NULL;
      }

      /*  check if the 'fh' is a link  */
      if(result_3->diropres_u.diropres.attributes.type == NFLNK) {

	 char *sym_lnk_data, *newName=NULL;

	 /*   read the link  */
	 if( (sym_lnk_data = nfs2ReadLink(hostname,
					  &result_3->diropres_u.diropres.file,
					  clnt))
	     == NULL) 
	   return NULL;

	 /*  if the symbolic link is wrt to the root dir, then the URL is 
	  *  redirected to that using the 'file' WWW protocol             */
	 if(sym_lnk_data[0] == '/') {
 	   extern char *redirecting_url;

	   StrAllocCopy(redirecting_url, "file://");
	   StrAllocCat (redirecting_url, hostname);
	   StrAllocCat (redirecting_url, sym_lnk_data);
	   errCode = HT_REDIRECTING;
	   return NULL;
	 }

	 if( (newName != NULL) && (*newName)) {
	   free(newName);
	   newName == NULL;
	 }

	 /*  create new filename */
         if( (newName = strdup(sym_lnk_data)) == NULL)
	     return NULL;
	 if(singlePath != NULL) {
	   StrAllocCat(newName, (singlePath +1) );
	   *singlePath = '/';
	 }
	 /*  update pointers  */
	 singlePath = filename = newName;
	 free(nfsproc_lookup_2_arg.name);
      }
      else {
	/* update the 'dir' handle */
	memcpy((void*) &nfsproc_lookup_2_arg.dir,
	       (void*) &result_3->diropres_u.diropres.file,
	       sizeof(nfs_fh3));
	if(singlePath != NULL) 
	  memcpy((void*)filename, (void*) singlePath, strlen(++singlePath)+1);
      }	/*  else  */
    } while (singlePath != NULL);
      
    free(nfsproc_lookup_2_arg.name);
 
    return result_3;
}


/*      Search the Prefix Cache
**	-----------------------
**
** On entry,
**	filename	address of a file/dir
** On exit,
**	fh              handle of the 'closest' filename found 
**                      NULL if not found
*/
PRIVATE char* SearchPrefixCache
ARGS2 (
  char *,	   filename,
  void **,         fh
)
{
  prefix *     tblptr;
  
  if(nfsCachePtr->prefixtable != NULL) {  
    int       tmplen, namelen = strlen(filename);
    int       retlen=0;
    prefix *  rettblptr = NULL;
    
    tblptr = nfsCachePtr->prefixtable;

    /* go thru the list */
    while(tblptr != NULL) {
      tmplen = strlen(tblptr->pathname);
      
      if(!strncmp(tblptr->pathname, filename, tmplen)) { /*  match found */
	/*  check for exact match. if so, update and return */
	if( namelen == tmplen ) {
	  *fh = (void*)tblptr->fh;
	  return &filename[namelen];
	}
	/*  check if this match is better than prev one and also consistent */
	else if( (tmplen > retlen)
		 && ( (tmplen == 1) || (filename[ tmplen ] == '/') ) ) {
	    rettblptr = tblptr;
	    retlen = tmplen;
	}
      }
      tblptr = tblptr->next;
    }/*  while */

    if(rettblptr != NULL) { /*  match found    */
      *fh = (void*) rettblptr->fh;
      return & ( filename[ strlen(rettblptr->pathname) ] );
    }
    else {                  /*  no match found */
      *fh = NULL;
      return filename;
    }
  }
  else /* nothing in table. return error */
      *fh = NULL;

    return filename;
}

/*      Insert in the Prefix Cache
**	--------------------------
**
** On entry,
**	filename	address of a file/dir
**	fh              handle of the 'closest' filename found 
** On exit,**
**                      
*/
PRIVATE void InsertPrefixCache
ARGS2 (
  char *,	   filename,
  void *,          fh
)
{
  prefix * tblptr = (prefix*) malloc(sizeof(prefix));

  if(tblptr == (prefix*) NULL)
    return;

  tblptr->next = NULL;
  
  if( (tblptr->pathname = strdup(filename)) == NULL) {
    free(tblptr);
    return;
  }

  /*  copy the file handle depending on the version  */
  if(nfsCachePtr->Version == NFS_V3) {
    /*  allocate entry and space for the handle */
    if( ((tblptr->fh = (void*)malloc(sizeof(nfs_fh3)))  == (void*) NULL)
       || (( ((nfs_fh3*)tblptr->fh)->data.data_val =
	                 (char*) malloc(((nfs_fh3*)fh)->data.data_len))
	   == NULL) ) {
      free(tblptr->pathname);
      free(tblptr);
    return;
    }

    ((nfs_fh3*)tblptr->fh)->data.data_len = ((nfs_fh3*)fh)->data.data_len ;
    /*  copy the handle  */
    memcpy( (void*) ((nfs_fh3*)tblptr->fh)->data.data_val,
	    (void*) ((nfs_fh3*)fh)->data.data_val,
	    ((nfs_fh3*)fh)->data.data_len );
  }
  else {   /*  Ver 2 */
    if( ((tblptr->fh = (void*)malloc(sizeof(nfs_fh)))  == (void*) NULL) ) {
      free(tblptr->pathname);
      free(tblptr);
    return;
    }
    memcpy(tblptr->fh, fh, sizeof(nfs_fh));
  }

  /*  update the table pointers  */
  if(nfsCachePtr->prefixtable == NULL)
    nfsCachePtr->prefixtable = tblptr;
  else {
    tblptr->next = nfsCachePtr->prefixtable;
    nfsCachePtr->prefixtable = tblptr;
  }

  return;
}

/*      Do Step Look Up
**	---------------
**
** On entry,
**	filename	address of a file/dir
** On exit,**
**     *Obj             handle of the filename
**     *Attr            its attributes
*/
BOOL nfs3StepLookUp
ARGS5 (
  char *,          hostname,
  char *,	   filename,
  nfs_fh3 **,      Obj,
  fattr3 **,       Attr,
  CLIENT *,        clnt
)
{
  char *        remainingPath;
  char *        name = strdup(filename);

  /* search thru the prefix table. if something found, look for the 
   * remaining part */
  remainingPath = SearchPrefixCache(filename, (void**)Obj);
  
  if(NFSDEBUGFLAG) fprintf(debugFile, "nfs3StepLookUp: remainpath: %s\n",
			    remainingPath);

  /*  nothing viable found. talk to the `mount daemon` */
  if(*Obj == NULL) {
    int status = nfs3Mount(hostname, filename, Obj, Attr, clnt);
    if(status)  /*  insert in the prefix table  */
      InsertPrefixCache(filename, *Obj);
    free(name);
    return status;
  }

  /*  lookup the remaining path  */
  if(*remainingPath)
  {
    LOOKUP3res* result_3 = nfs3singleLookUp(hostname, remainingPath, *Obj, clnt);

    if(result_3 == NULL) {
       if(FirstTime && (nfs_error == NFS_ERR_RPC) )
         return 0;
       else {
	 int status = nfs3Mount(hostname, name, Obj, Attr, clnt);
	 if(status)  /*  insert in the prefix table  */
	   InsertPrefixCache(name, *Obj);
	 free(name);
	 return status;
       }
    }
    
    *Obj  = &result_3->LOOKUP3res_u.resok.object;
    *Attr = &result_3->LOOKUP3res_u.resok.obj_attributes.post_op_attr_u.attributes;
  }

  if(*remainingPath)  /*  update the prefix table  */
    InsertPrefixCache(name, *Obj);
  else {              /*  we found the exact entry. so do a 'getattr' to get 
                       *  the attributes           */
    GETATTR3res     *result_2;
    GETATTR3args     nfsproc3_getattr_3_arg;

    free(name);
    memcpy( (void*) &nfsproc3_getattr_3_arg.object,
	    (void*) *Obj,
	    sizeof(nfs_fh3) );
    
    result_2 = nfsproc3_getattr_3(&nfsproc3_getattr_3_arg, clnt);
    if (result_2 == (GETATTR3res *) NULL) {
      struct rpc_err rpcErr;

      if(NFSDEBUGFLAG) clnt_perror(clnt, "call failed");

      clnt_geterr(clnt, &rpcErr);
      if(rpcErr.re_status == RPC_CANTRECV) 
	nfs_error = NFS_ERR_RPC;
     
      return 0;
    }
    if(result_2->status != NFS3_OK)
      return 0;

    *Attr = &result_2->GETATTR3res_u.resok.obj_attributes;
  }

  free(name);
  return 1;
}

/*      Do Step Look Up
**	---------------
**
** On entry,
**	filename	address of a file/dir
** On exit,**
**     *Obj             handle of the filename
**     *Attr            its attributes
*/
BOOL nfs2StepLookUp
ARGS5 (
  char *,          hostname,
  char *,	   filename,
  nfs_fh **,       Obj,
  fattr **,        Attr,
  CLIENT *,        clnt
)
{
  char *        remainingPath;
  char *        name = strdup(filename);

  /* search thru the prefix table. if something found, look for the 
   * remaining part */
  remainingPath = SearchPrefixCache(filename, (void**)Obj);
  
  if(NFSDEBUGFLAG) fprintf(debugFile, "nfs2StepLookUp: remainpath: %s\n",
			    remainingPath);

  /*  nothing viable found. talk to the `mount daemon` */
  if(*Obj == NULL) {
    int status = nfs2Mount(hostname, filename, Obj, Attr, clnt);
    if(status)  /*  insert in the prefix table  */
      InsertPrefixCache(filename, *Obj);
    free(name);
    return status;
  }

  /*  lookup the remaining path  */
  if(*remainingPath)
  {
    diropres* result_3 = nfs2singleLookUp(hostname, remainingPath, *Obj, clnt);

    if(result_3 == NULL) {
      if(FirstTime && (nfs_error == NFS_ERR_RPC) )
	return 0;
      else {
	int status = nfs2Mount(hostname, name, Obj, Attr, clnt);
	if(status)  /*  insert in the prefix table  */
	  InsertPrefixCache(name, *Obj);
	free(name);
	return status;
      }
    }
    
    *Obj  = &result_3->diropres_u.diropres.file;
    *Attr = &result_3->diropres_u.diropres.attributes;
  }

  if(*remainingPath)  /*  update the prefix table  */
    InsertPrefixCache(name, *Obj);
  else {              /*  we found the exact entry. so do a 'getattr' to get 
                       *  the attributes           */
    attrstat          *result_2;
    static nfs_fh     nfsproc_getattr_2_arg;

    free(name);
    memcpy( (void*) &nfsproc_getattr_2_arg,
	    (void*) *Obj,
	    sizeof(nfs_fh) );
    
    result_2 = nfsproc_getattr_2(&nfsproc_getattr_2_arg, clnt);
    if (result_2 == (attrstat *) NULL) {
      struct rpc_err rpcErr;

      if(NFSDEBUGFLAG) clnt_perror(clnt, "call failed");

      clnt_geterr(clnt, &rpcErr);
      if(rpcErr.re_status == RPC_CANTRECV) 
	nfs_error = NFS_ERR_RPC;
     
      return 0;
    }
    if(result_2->status != NFS3_OK)
      return 0;

    *Attr = &result_2->attrstat_u.attributes;
  }

  free(name);
  return 1;
}


/*	Talk to the Mount daemon and Get the 'fh' for the 'addr'
**	-------------------------------------------------------
**
** On entry,
**	filename	addr of the document
** On exit,
**	*Object         'fh' of the document
**      *Attr           attributes of the document 
**			0 if bad.
*/

BOOL nfs3Mount
ARGS5 (
  char *,                      hostname,
  char *,                      filename,
  nfs_fh3 **,                  Object,
  fattr3 **,                   Attr,
  CLIENT *,                    clnt
)
{
  mountres3  *result_m;
  dirpath     mountproc_mnt_3_arg;
  char       *strport = NULL;

  if(NFSDEBUGFLAG) fprintf(debugFile, "nfs3Mount:  trying %s\n", filename);

  /*  check for interruption  */
  if(HTCheckActiveIcon(0)) {
    HTProgress("Data transfer interrupted.");
    return 0;
  }

  if(strport = strchr(hostname, ':')) 
    *strport++ = 0;                  /* Chop off port */

  /*  create the client  */
  if( (nfsCachePtr->MountClnt 
       = clnt_create(hostname, MOUNTPROG, nfsCachePtr->Version, "udp"))
      == (CLIENT*) NULL)
    return 0;
  nfsCachePtr->MountClnt->cl_auth = authsys_create_default();

  if( (mountproc_mnt_3_arg = strdup(filename)) == NULL) {
    clnt_destroy(nfsCachePtr->MountClnt);
    return 0;
  }

  /*  talk to the daemon  */
  result_m = mountproc_mnt_3(&mountproc_mnt_3_arg, nfsCachePtr->MountClnt);
  if (result_m == (mountres3 *) NULL) {
    if(NFSDEBUGFLAG) clnt_perror(clnt, "call failed");
    clnt_destroy(nfsCachePtr->MountClnt);
    free(mountproc_mnt_3_arg);
    nfs_error = NFS_ERR_MNT;
    return 0;
  }
  if(result_m->fhs_status != MNT_OK) {
    clnt_destroy(nfsCachePtr->MountClnt);
    free(mountproc_mnt_3_arg);
    nfs_error = NFS_ERR_MNT;
    return 0;
  }

  mountproc_umnt_3(&mountproc_mnt_3_arg, nfsCachePtr->MountClnt);
  clnt_destroy(nfsCachePtr->MountClnt);

  /*  get the attributes  */
  {
    GETATTR3res          *result_2;
    static GETATTR3args  nfsproc3_getattr_3_arg;
    
    /*  check for interruption  */
    if(HTCheckActiveIcon(0)) {
      HTProgress("Data transfer interrupted.");
      return 0;
    }

    memcpy( (void*) &nfsproc3_getattr_3_arg.object,
	    (void*) &result_m->mountres3_u.mountinfo.fhandle,
	    sizeof(nfs_fh3) );
    
    result_2 = nfsproc3_getattr_3(&nfsproc3_getattr_3_arg, clnt);
    if (result_2 == (GETATTR3res *) NULL) {
      struct rpc_err rpcErr;

      if(NFSDEBUGFLAG) clnt_perror(clnt, "call failed");

      clnt_geterr(clnt, &rpcErr);
      if(rpcErr.re_status == RPC_CANTRECV) 
	nfs_error = NFS_ERR_RPC;
     
      return 0;
    }
    if(result_2->status != NFS3_OK)
      return 0;

    *Object = &nfsproc3_getattr_3_arg.object;
    *Attr   = &result_2->GETATTR3res_u.resok.obj_attributes;
  }

  if(NFSDEBUGFLAG)  fprintf(debugFile, "........................successful!\n");
  return 1;
}


/*	Talk to the Mount daemon and Get the 'fh' for the 'addr'
**	-------------------------------------------------------
**
** On entry,
**	filename	addr of the document
** On exit,
**	*Object         'fh' of the document
**      *Attr           attributes of the document 
**			0 if bad.
*/

BOOL nfs2Mount
ARGS5 (
  char *,                      hostname,
  char *,                      filename,
  nfs_fh **,                   Object,
  fattr **,                    Attr,
  CLIENT *,                    clnt
)
{
  fhstatus  *result_m;
  dirpath    mountproc_mnt_1_arg;
  char      *strport = NULL;

  if(NFSDEBUGFLAG) fprintf(debugFile, "nfs2Mount: trying  %s\n", filename);

  /*  check for interruption  */
  if(HTCheckActiveIcon(0)) {
    HTProgress("Data transfer interrupted.");
    return 0;
  }

  if(strport = strchr(hostname, ':')) 
    *strport++ = 0;                  /* Chop off port */

  /*  create the client  */
  if( (nfsCachePtr->MountClnt 
       = clnt_create(hostname, MOUNTPROG, MOUNTVERS, "udp"))
      == (CLIENT*) NULL)
    return 0;

  if( (mountproc_mnt_1_arg = strdup(filename)) == NULL) {
    clnt_destroy(nfsCachePtr->MountClnt);
    return 0;
  }

  /*  talk to the daemon  */
  result_m = mountproc_mnt_1(&mountproc_mnt_1_arg, nfsCachePtr->MountClnt);
  if (result_m == (fhstatus *) NULL) {
    if(NFSDEBUGFLAG) clnt_perror(clnt, "call failed");
    free(mountproc_mnt_1_arg);
    clnt_destroy(nfsCachePtr->MountClnt);
    nfs_error = NFS_ERR_MNT;
    return 0;
  }
  if(result_m->fhs_status != MNT_OK) {
    free(mountproc_mnt_1_arg);
    clnt_destroy(nfsCachePtr->MountClnt);
    nfs_error = NFS_ERR_MNT;
    return 0;
  }
 
  mountproc_umnt_1(&mountproc_mnt_1_arg, nfsCachePtr->MountClnt);
  clnt_destroy(nfsCachePtr->MountClnt);

  /*  get the attributes  */
  {
    attrstat          *result_2;
    static nfs_fh     nfsproc_getattr_2_arg;
    
    /*  check for interruption  */
    if(HTCheckActiveIcon(0)) {
      HTProgress("Data transfer interrupted.");
      return 0;
    }

    memcpy( (void*) &nfsproc_getattr_2_arg,
	    (void*) &result_m->fhstatus_u.fhs_fhandle,
	    sizeof(nfs_fh) );
    
    result_2 = nfsproc_getattr_2(&nfsproc_getattr_2_arg, clnt);
    if (result_2 == (attrstat*) NULL) {
      struct rpc_err rpcErr;

      if(NFSDEBUGFLAG) clnt_perror(clnt, "call failed");

      clnt_geterr(clnt, &rpcErr);
      if(rpcErr.re_status == RPC_CANTRECV) 
	nfs_error = NFS_ERR_RPC;
     
      return 0;
    }
    if(result_2->status != NFS3_OK)
      return 0;

    *Object = &nfsproc_getattr_2_arg;
    *Attr   = &result_2->attrstat_u.attributes;
  }

  if(NFSDEBUGFLAG) fprintf(debugFile, "..................successful!\n");
  return 1;
}

/*	Creates connection to arbitrary port and then a Client handle
**	-------------------------------------------------------------
**
** On entry,
**	hostname        name of the host
**      port            port!
** On exit,
**      NFSClientPtr    fills in the Client handle and socket descriptor
*/
PRIVATE BOOL CreatePortConnection
ARGS5 (
  char *,                      hostname,
  int,                         port,
  int,                         prog,
  struct netconfig,            *nconf,
  nfsCache **,                 NFSCachePtr
)
{
        nfsCache *          CachePtr = *NFSCachePtr;
	CLIENT *            clnt;
        struct netbuf *     nb = NULL;
        struct t_bind *     tbind = NULL;
        struct timeval      tv;
	struct t_info       tinfo;
        int                 fd = -1;
	int                 vers;

        if (nconf == NULL) 
	  return (NULL);

        if ((fd = t_open(nconf->nc_device, O_RDWR, &tinfo)) == -1) 
	  return 0;

        /* LINTED pointer alignment */
        if ((tbind = (struct t_bind *) t_alloc(fd, T_BIND, T_ADDR))
	    == NULL) {
	  t_close(fd);
	  return 0;
        }

	clnt = CachePtr->clnt;

	/* loop till connection is made      */
	for(vers=NFS_V3; vers >= NFS_VERSION ;vers--) {
	  if (rpcb_getaddr(prog, vers, nconf, &tbind->addr, hostname) == 0)
	    continue;
 
	  /* LINTED pointer alignment */
	  ((struct sockaddr_in *) tbind->addr.buf)->sin_port = htons(port);
 
	  clnt = clnt_tli_create(fd, nconf, &tbind->addr, prog, vers, 0, 0);
	  if (clnt == NULL) 
	    continue;

	  /* check the client handle */
	  tv.tv_sec = 5;
	  tv.tv_usec = 0;
	  if( clnt_call(clnt, NULLPROC, xdr_void, 0, xdr_void, 0, tv)
	      != RPC_SUCCESS) {
	    clnt_destroy(clnt);
	    continue;
	  }

	  /* fill in the req info */
	  CachePtr->Version = vers;
	  CachePtr->fd      = fd;
	  CachePtr->tbind   = tbind;
	  return 1;
	}

	t_close(fd);
	t_free((char *) tbind, T_BIND);
	return 0;
}

/*	Creates A Client on an arbitrary port
**	-------------------------------------
**
** On entry,
**	hostname        name of the host
**      port            port!
** On exit,
**      returns         1 if succesful
**                      0 if not
*/
BOOL CreatePortClient
ARGS4 (
  char *,                      hostname,
  int,                         port,
  int,                         prog,
  nfsCache **,                 NFSCachePtr
)
{
        struct netbuf *nb = NULL;
        struct netconfig *nconf;
        NCONF_HANDLE *nc = NULL;
        int found_tcp = FALSE;
 
	/*  initialize  */
	nc = setnetpath();
	if (nc == NULL)
	  return 0;

	/* check for tcp connection first */
	while (nconf = getnetpath(nc)) {
	    if (strcmp(nconf->nc_netid, "tcp") == 0) {
	      if ((strcmp(nconf->nc_protofmly, NC_INET) == 0) &&
		  ((strcmp(nconf->nc_proto, NC_TCP) == 0))) {
		found_tcp = TRUE;
		break;
	      }
	    }
	}

	if(!found_tcp && setnetpath()) {
	  while (nconf = getnetpath(nc)) {
	    if (strcmp(nconf->nc_netid, "udp") == 0) {
	      if ((strcmp(nconf->nc_protofmly, NC_INET) == 0) &&
		  ((strcmp(nconf->nc_proto, NC_UDP) == 0))) {
		break;
	      }
	      else 
		return 0;
	    }
	  }
	}

        /* Create the connection based on the given prog/vers/port/nconf */
        if (!(CreatePortConnection(hostname, port, prog, nconf, NFSCachePtr))) {
	  if (nc)
	    endnetconfig(nc);
	  return 0;
	}

        if(nc)
	  endnetconfig(nc);
	
        return 1;
}

