/*
 * @(#)v3.x	1.20 03 Feb 1994
 *
 * This file defines version 3 of the NFS protocol.
 * This protocol must be used in conjunction with
 * version 3 of the MOUNT protocol and version 4
 * of the NLM protocol.
 */

const NFS3_FHSIZE = 64;
const NFS3_COOKIEVERFSIZE = 8;
const NFS3_CREATEVERFSIZE = 8;
const NFS3_WRITEVERFSIZE = 8;

typedef unsigned hyper	uint64;
typedef hyper		int64;
typedef unsigned long	uint32;
typedef long		int32;
typedef string		filename3<>;
typedef string		nfspath3<>;
typedef uint64		fileid3;
typedef uint64		cookie3;
typedef uint32		uid3;
typedef uint32		gid3;
typedef uint64		size3;
typedef uint64		offset3;
typedef uint32		mode3;
typedef uint32		count3;
typedef opaque		cookieverf3[NFS3_COOKIEVERFSIZE];
typedef opaque		createverf3[NFS3_CREATEVERFSIZE];
typedef opaque		writeverf3[NFS3_WRITEVERFSIZE];

/*
 * Fhandle
 */
struct nfs_fh3 {
	opaque		data<NFS3_FHSIZE>;
};

/*
 * Diropargs
 */
struct diropargs3 {
	nfs_fh3		dir;
	filename3	name;
};

/*
 * Timeval
 */
struct nfstime3 {
	uint32		seconds;	/* seconds */
	uint32		nseconds;	/* nanoseconds */
};

/*
 * Specdata
 *
 * The interpretation of the two words depends on the type of
 * file system object.  For a block special (NF3BLK) or character
 * special (NF3CHR), the words are the major and minor device
 * numbers.  (This is obviously a Unix-specific interpretation.)
 *
 * For other types (particularly NF3REG and NF3DIR), there are two
 * possibilities:
 * - No special data.  In this case, both words are 0.  This
 *   is the default case.
 * - 'specdata1' contains a 4 byte space filled vendor code.  In
 *   this case, 'specdata2' contains an opaque vendor-specific
 *   value.
 *
 * [The vendor code should consist of four ASCII characters, stored
 * in network order.  The preferred code is the vendor stock
 * quotation symbol, in upper-case - for example, "SUNW", or "IBM ".
 * For other companies, a code should be derived from the vendor
 * name published in the "Assigned Numbers" RFC, and should use
 * lower-case letters.  Experimental or site-specific codes should
 * consist of numerals and spaces only.]
 *
 * A server vendor may use this mechanism to convey a single
 * word of vendor-specific information about a file.  For example,
 * a vendor may choose to indicate that the corresponding file
 * has been migrated to a secondary storage medium.  A client
 * which was aware of the encoding might provide this information
 * as part of an expanded directory listing.
 *
 * An NFS V3 server should not be written in such a way that
 * correct operation relies upon the client's interpretation of
 * this information.  Furthermore, it should be emphasized that the
 * NFS V3 protocol does not include any features which use this
 * information in any way.  If more complex vendor-specific services
 * are required, they should be provided using a distinct RPC
 * protocol.  Pursuing the earlier example, a vendor might provide
 * a specific protocol whereby the client could interrogate
 * and control a file migration mechanism on the server.
 */
struct specdata3 {
	uint32	specdata1;	/* major device OR 0 OR vendor code */
	uint32	specdata2;	/* minor device OR 0 OR vendor-specific value */
};

/*
 * Error status.
 */
enum nfsstat3 {
	NFS3_OK= 0,		/* no error */
	NFS3ERR_PERM=1,		/* Not owner */
	NFS3ERR_NOENT=2,	/* No such file or directory */
	NFS3ERR_IO=5,		/* I/O error */
	NFS3ERR_NXIO=6,		/* No such device or address */
	NFS3ERR_ACCES=13,	/* Permission denied */
	NFS3ERR_EXIST=17,	/* File exists */
	NFS3ERR_XDEV=18,	/* Attempted cross-device link */
	NFS3ERR_NODEV=19,	/* No such device */
	NFS3ERR_NOTDIR=20,	/* Not a directory*/
	NFS3ERR_ISDIR=21,	/* Is a directory */
	NFS3ERR_INVAL=22,	/* Invalid argument */
	NFS3ERR_FBIG=27,	/* File too large */
	NFS3ERR_NOSPC=28,	/* No space left on device */
	NFS3ERR_ROFS=30,	/* Read-only file system */
	NFS3ERR_MLINK=31,	/* Too many links */
	NFS3ERR_NAMETOOLONG=63,	/* File name too long */
	NFS3ERR_NOTEMPTY=66,	/* Directory not empty */
	NFS3ERR_DQUOT=69,	/* Disc quota exceeded */
	NFS3ERR_STALE=70,	/* Stale NFS file handle */
	NFS3ERR_REMOTE=71,	/* Too many levels of remote in path */
	NFS3ERR_BADHANDLE=10001, /* Illegal NFS file handle */
	NFS3ERR_NOT_SYNC=10002,	/* update synch mismatch */
	NFS3ERR_BAD_COOKIE=10003, /* readdir cookie is stale */
	NFS3ERR_NOTSUPP=10004,	/* operation not supported */
	NFS3ERR_TOOSMALL=10005,	/* buffer/request too small */
	NFS3ERR_SERVERFAULT=10006, /* An i/o or similar failure caused */
				/* the server to abandon the request */
				/* No attributes can be returned.  The */
				/* client should translate this into EIO */
	NFS3ERR_BADTYPE=10007,	/* Attempt to CREATE an object of a */
				/* type not supported by the server. */
	NFS3ERR_JUKEBOX=10008	/* server can't process request now, client */
				/* should wait and then try the request */
				/* with a new xid */
};

/*
 * File types.
 */
enum ftype3 {
	NF3REG =  1,		/* regular file */
	NF3DIR =  2,		/* directory */
	NF3BLK =  3,		/* block special */
	NF3CHR =  4,		/* character special */
	NF3LNK =  5,		/* symbolic link */
	NF3SOCK = 6,		/* unix domain sockets */
	NF3FIFO = 7		/* named pipe */
};

/*
 * File attributes.
 *
 * This structure defines the attributes of a file system object.
 * It is returned by most operations on an object; in the case
 * of operations which affect two objects, the attributes for
 * both may be returned.  In some cases, the attributes are
 * returned in a structure "wcc_data" defined below; in other
 * cases the attributes are returned alone.  The choice depends
 * on the requirements of the attribute cache management software.
 *
 * The main changes from NFS V2 are that many of the
 * fields have been widened, and the major/minor device
 * information is now presented in a distinct structure
 * rather than being packed into a word.
 */
struct fattr3 {
	ftype3		type;	/* file type */
	mode3		mode;	/* protection mode bits */
	uint32		nlink;	/* # hard links */
	uid3		uid;	/* owner user id */
	gid3		gid;	/* owner group id */
	size3		size;	/* file size in bytes */
	size3		used;	/* space used by file in bytes */
	specdata3	rdev;	/* special file info */
	uint64		fsid;	/* device # */
	fileid3		fileid;	/* inode # */
	nfstime3	atime;	/* time of last access */
	nfstime3	mtime;	/* time of last modification */
	nfstime3	ctime;	/* time of last change */
};

/*
 * The structure used for returning attributes for operations which
 * are not directly involved with manipulating attributes.
 *
 * A general notion for this version of the protocol is to return
 * the attributes on every operation.  This is designed to help
 * clients which choose to cache attributes by reducing the number
 * of calls to GETATTR which are only done to check to make sure
 * that the cached attributes are valid.
 *
 * This appears to make returning attributes optional.  This is
 * true in fact, however, server implementors are *strongly*
 * encouraged to make best effort to return attributes.
 */
union post_op_attr switch (bool attributes_follow) {
case TRUE:
	fattr3	attributes;
case FALSE:
	void;
};

/*
 * Subset of pre-operation attributes needed for weak cache consistency.
 *
 * When a client performs an operation which modifies the state of
 * a file or directory on the server, it cannot immediately determine
 * from the post-operation attributes whether the operation just
 * performed was the only operation on the object since the last
 * time it received the attributes for the object.  This is important,
 * since if an intervening operation *has* changed the object, the
 * client should invalidate any cached data for the object (except for
 * the data which it just wrote).
 */
struct wcc_attr {
	size3		size;	/* file size in bytes */
	nfstime3	mtime;	/* time of last modification */
	nfstime3	ctime;	/* time of last change */
};

/*
 * To deal with this, we introduce the notion of "weak cache consistency
 * data", or "wcc_data".  A wcc_data structure consists of certain
 * key fields from the object attributes *before* the operation,
 * together with the object attributes *after* the operation.
 * This information allows the client to manage its cache more
 * accurately than in NFS V2.  (The term "weak cache consistency" is
 * used to emphasize the fact that this mechanism does not provide
 * the degree of consistency which a full-blown cache consistency
 * protocol would provide.)
 */
union pre_op_attr switch (bool attributes_follow) {
case TRUE:
	wcc_attr	attributes;
case FALSE:
	void;
};

struct wcc_data {
	pre_op_attr	before;
	post_op_attr	after;
};

/*
 * One of the principles of this revision of the NFS protocol is to
 * return the real value from the indicated operation and not an error
 * from an incidental operation.  Such errors can result from attempting
 * to get the file attributes and from attempting to construct a file
 * handle for a given object.  The post_op_attr structure was designed
 * to allow the server to recover from getattr-type errors.  This
 * strucuture is designed to allow the server to recover from errors
 * encountered when trying to construct a file handle.
 *
 * This structure is used to return a file handle from the CREATE,
 * MKDIR, SYMLINK, and MKNOD requests.  In each case, the client can
 * get the file handle by issuing a LOOKUP request after a successful
 * return from one of the above operations.  Returning the file handle
 * is an optimization to not force the client to immediately issue a
 * LOOKUP request.
 */
union post_op_fh3 switch (bool handle_follows) {
case TRUE:
	nfs_fh3		handle;
case FALSE:
	void;
};

/*
 * Settable file attributes.
 *
 * In NFS V2, the settable attributes were represented by
 * a subset of the file attributes structure; the client
 * indicated those attributes which were *not* to be
 * updated by setting the corresponding field to -1.
 *
 * In NFS V3, the settable attributes are described by a
 * structure containing a set of discriminated unions.  Each
 * union indicates whether the corresponding attribute
 * is to be updated, and if so how.
 *
 * There are two forms of union used.
 *    In setting the mode, UID, GID or size, the discriminated
 *    union is switched on a boolean 'set_it'; if this is TRUE,
 *    a value of the appropriate type is then encoded.
 *
 *    In setting the atime or mtime, the union is switched on
 *    an enumerated 'time_how set_it'.  If set_it has the value
 *    DONT_CHANGE, the corresponding attribute is left unchanged.
 *    If it has the value SET_TO_SERVER_TIME, the corresponding
 *    attribute is set by the server to its local time; no data
 *    is provided by the client.  Finally, if set_it is
 *    SET_TO_CLIENT_TIME, the attribute is set to the time
 *    passed by the client in a struct nfstime3.
 *    (See also the description of the FSINFO procedure, which
 *    addresses the issue of time granularity.)
 */
enum time_how {
	DONT_CHANGE = 0,
	SET_TO_SERVER_TIME = 1,
	SET_TO_CLIENT_TIME = 2
};
union set_mode3 switch (bool set_it) {
case TRUE:
	mode3		mode;
default:
	void;
};
union set_uid3 switch (bool set_it) {
case TRUE:
	uid3		uid;
default:
	void;
};

union set_gid3 switch (bool set_it) {
case TRUE:
	gid3		gid;
default:
	void;
};

union set_size3 switch (bool set_it) {
case TRUE:
	size3		size;
default:
	void;
};

union set_atime switch (time_how set_it) {
case SET_TO_CLIENT_TIME:
	nfstime3	atime;
default:
	void;
};

union set_mtime switch (time_how set_it) {
case SET_TO_CLIENT_TIME:
	nfstime3	mtime;
default:
	void;
};

struct sattr3 {
	set_mode3	mode;	/* protection mode bits */
	set_uid3	uid;	/* owner user id */
	set_gid3	gid;	/* owner group id */
	set_size3	size;	/* file size in bytes */
	set_atime	atime;	/* time of last access */
	set_mtime	mtime;	/* time of last modification */
};

/*
 * File name component handling.
 *
 * Server implementations of NFS V3 procedures which create directory
 * entries (CREATE, MKDIR, SYMLINK, MKNOD, RENAME and LINK) will
 * frequently impose restrictions on the names which can be created.
 * Most will disallow the use of "." and "..".  Many servers will also
 * forbid the use of names which contain certain characters, such as
 * the path component separator used by the server operating system.
 * The exact error status values return for these errors is specified
 * in the description of each procedure argument; the reader is
 * cautioned that the values (which conform to NFS V2 server practice)
 * are not necessarily obvious, nor are they consistent from one
 * procedure to the next.
 *
 * The following comments apply to all NFS V3 procedures in which
 * the client provides one or more file names in the arguments:
 * LOOKUP, CREATE, MKDIR, SYMLINK, MKNOD, REMOVE, RMDIR, RENAME, and
 * LINK.
 *
 *    (i) If the name is null (i.e. zero length), the procedure
 *        will fail with NFS3ERR_ACCES.
 *   (ii) If the name is longer than 'name_max' (see PATHCONF),
 *        then if 'no_trunc' is TRUE, the procedure will fail with
 *        NFS3ERR_NAMETOOLONG.  Otherwise the name will be truncated
 *        to 'name_max' bytes.
 */

/*
 * NFS V3 procedure NFSPROC3_NULL (index 0).
 */

/*
 * NFS V3 procedure NFSPROC3_GETATTR (index 1).
 *
 * The function of procedure GETATTR is to retrieve
 * the attributes for a specified file system object.
 * The object is identified by the file handle which
 * the server returned as part of a LOOKUP, CREATE, MKDIR,
 * SYMLINK, MKNOD, or READDIRPLUS procedure (or from the
 * mount service, described elsewhere).
 *
 * On entry, the arguments in GETATTR3args are as follows:
 *
 *	object		The file handle for the object.
 *
 * On a successful return, the results in GETATTR3resok
 * are as follows:
 *
 *	obj_attributes	The fattr3 structure containing
 *			the object attributes.
 *
 * On any error, no results are returned.
 *
 * The following structure defines the arguments to the
 * procedure GETATTR:
 */
struct GETATTR3args {
	nfs_fh3		object;
};
/*
 * The following union defines the results from invoking the
 * procedure GETATTR.
 */
struct GETATTR3resok {
	fattr3		obj_attributes;
};

union GETATTR3res switch (nfsstat3 status) {
case NFS3_OK:
	GETATTR3resok	resok;
default:
	void;
};

/*
 * NFS V3 procedure NFSPROC3_SETATTR (index 2).
 *
 * The function of procedure SETATTR is to change one or
 * more of the attributes of an file system object on the
 * server.  The attributes to be changed, and their new
 * values, are specified by a struct sattr3 (described
 * above).
 *
 * Note that not all attributes are applicable to all types
 * of file system objects.  An attempt to set an inappropriate
 * attribute for the object results in XXX.
 *
 * A client may request that the server check that the
 * object in in an expected state before performing
 * the SETATTR operation.  To do this, it sets the argument
 * 'guard.check' to TRUE and passes a time value in
 * 'guard.obj_ctime'.  If 'guard.check' is TRUE, the server
 * must compare the value of 'guard.obj_ctime' to the current
 * 'ctime' of the object.  If the values are not equal, the server
 * must not change the object attributes and must return a status
 * of NFS3ERR_NOT_SYNC.
 *
 * If 'guard.check' is FALSE, the server will not perform this
 * check.
 *
 * This mechanism allows the client to avoid changing the
 * attributes of an object on the basis of stale attributes.
 * It must be emphasized that this does NOT provide exactly-once
 * semantics for this operation.  In particular, if a reply is lost
 * and the server's duplicate request cache does not detect
 * the retransmission of the request, the procedure can fail with
 * NFS3ERR_NOT_SYNC even though the attribute setting was
 * performed successfully.
 *
 * On entry, the arguments in SETATTR3args are as follows:
 *
 *	object		The file handle for the object.
 *	new_attributes	A sattr3 structure describing the
 *			attributes to be changed and the
 *			new values for those attributes.
 *	guard		A sattrguard3 union.
 *
 * On a successful return, the results in SETATTR3resok
 * are as follows:
 *
 *	obj_wcc		A wcc_data structure containing the
 *			old and new attributes for the object.
 *
 * On certain errors (as shown in the discriminated union
 * SETATTR3res below) the following data is returned in the
 * structure SETATTR3resfail:
 *
 *	obj_wcc		A wcc_data structure containing the
 *			old and new attributes for the object.
 *
 * On other errors, no results are returned.
 *
 * The following structures define the arguments to the
 * procedure SETATTR:
 */
union sattrguard3 switch (bool check) {
case TRUE:
	nfstime3	obj_ctime;
case FALSE:
	void;
};

struct SETATTR3args {
	nfs_fh3		object;
	sattr3		new_attributes;
	sattrguard3	guard;
};
/*
 * The following union defines the results from invoking the
 * procedure SETATTR.
 */
struct SETATTR3resok {
	wcc_data	obj_wcc;
};

struct SETATTR3resfail {
	wcc_data	obj_wcc;
};

union SETATTR3res switch (nfsstat3 status) {
case NFS3_OK:
	SETATTR3resok	resok;
default:
	SETATTR3resfail	resfail;
};

/*
 * NFS V3 procedure NFSPROC3_LOOKUP (index 3).
 *
 * The function of procedure LOOKUP is to search a directory for
 * a specific name and return the attributes and file handle
 * for the corresponding file system object.  If the search
 * is successful, the status NFS3_OK is returned.  If the name is
 * not found in the directory, the status NFS3ERR_NOENT is returned.
 *
 * If the file name is longer than the maximum defined for the
 * file system (see PATHCONF 'name_max') and the file system
 * does not truncate names (see PATHCONF 'no_trunc'), the status
 * NFS3ERR_NAMETOOLONG is returned.
 *
 * If 'what.dir' does not identify a directory, the status
 * NFS3ERR_NOTDIR is returned.
 *
 * Two names have special significance.  The name "." is to be
 * interpreted as referring to the directory itself, while ".."
 * is to be interpreted as referring to the "parent" directory; i.e.
 * the directory which includes the specified directory as a
 * member.
 *
 * On entry, the arguments in LOOKUP3args are as follows:
 *
 *	what.dir	The file handle for the directory to
 *			be searched.
 *	what.name	The name to be searched for.
 *
 * On a successful return, the results in LOOKUP3resok
 * are as follows:
 *
 *	object		The file handle for the object
 *			corresponding to what.name.
 *	obj_attributes	The attributes for the object 'object'.
 *	dir_attributes	The attributes for the directory
 *			'what.dir'.
 *
 * On certain errors (as shown in the discriminated union
 * LOOKUP3res below) the following data is returned in the
 * structure LOOKUP3resfail:
 *
 *	dir_attributes	The attributes for the directory
 *			what.dir.
 *
 * On other errors, no results are returned.
 *
 * The following structure defines the arguments to the
 * procedure LOOKUP:
 */
struct LOOKUP3args {
	diropargs3	what;
};
/*
 * The following union defines the results from invoking the
 * procedure LOOKUP.
 */
struct LOOKUP3resok {
	nfs_fh3		object;
	post_op_attr	obj_attributes;
	post_op_attr	dir_attributes;
};

struct LOOKUP3resfail {
	post_op_attr	dir_attributes;
};

union LOOKUP3res switch (nfsstat3 status) {
case NFS3_OK:
	LOOKUP3resok	resok;
default:
	LOOKUP3resfail	resfail;
};

/*
 * NFS V3 procedure NFSPROC3_ACCESS (index 4).
 *
 * The function of procedure ACCESS is to determine the access rights
 * which a user has with respect to a file or directory.  In general,
 * it is insufficient for the client to attempt to deduce this
 * information by inspecting the 'uid', 'gid' and 'mode' fields in the
 * file attributes, since the server may perform uid or gid mapping or
 * enforce additional access control restrictions.  In NFS V2, the only
 * reliable way to determine whether an operation was allowed was to
 * try it and see if it succeeded or failed.  Using the ACCESS
 * procedure, the client can ask the server to indicate whether or not
 * one or more classes of operations are permitted.
 *
 * The client encodes the set of permissions which are to be checked in
 * a bitmask.  The server checks the permissions encoded in the bitmask
 * and returns a bitmask encoded with the permissions that the client
 * is allowed.  The server should return a status of NFS3_OK if no
 * errors occurred or if the only error that occurred during processing
 * was a "permission denied" error from doing one or more of the access
 * checks.
 *
 * It should be emphasized that the results of this procedure are
 * necessarily advisory in nature.
 *
 * On entry, the arguments in ACCESS3args are as follows:
 *
 *	object		The file handle for the file system object to
 *			which access is to be checked.
 *	access		A bit mask of access permissions.
 *
 * On a successful return, the results in ACCESS3resok are as follows:
 *
 *	obj_attributes	The attributes of 'object'.
 *	access		A bit mask of access permissions.
 *
 * On certain errors (as shown in the discriminated union ACCESS3res
 * below) the following data is returned in the structure
 * ACCESS3resfail:
 *
 *	obj_attributes	The attributes of 'object'.
 *
 * On other errors, no results are returned.
 *
 * The following structure defines the arguments to the procedure
 * ACCESS:
 */
struct ACCESS3args {
	nfs_fh3		object;
	uint32		access;	/* bit mask - see below */
};

/*
 * The following access permissions may be requested:
 */
const ACCESS3_READ = 0x1;	/* read data or readdir a directory */
const ACCESS3_LOOKUP = 0x2;	/* lookup a name in a directory */
const ACCESS3_MODIFY = 0x4;	/* rewrite existing file data or */
				/* modify existing directory entries */
const ACCESS3_EXTEND = 0x8;	/* write new data or add directory entries */
const ACCESS3_DELETE = 0x10;	/* delete existing directory entry */
const ACCESS3_EXECUTE = 0x20;	/* execute file (no meaning for a directory) */

/*
 * The following union defines the results from invoking the
 * procedure ACCESS.
 */
struct ACCESS3resok {
	post_op_attr	obj_attributes;
	uint32		access;	/* bit mask - see above */
};

struct ACCESS3resfail {
	post_op_attr	obj_attributes;
};

union ACCESS3res switch (nfsstat3 status) {
case NFS3_OK:
	ACCESS3resok	resok;
default:
	ACCESS3resfail	resfail;
};

/*
 * NFS V3 procedure NFSPROC3_READLINK (index 5).
 *
 * The function of procedure READLINK is to read the data
 * associated with a symbolic link.  The argument 'symlink' must
 * identify an object of type NF3LNK.  If it does not, the status
 * NFS3ERR_INVAL is returned.  (Note: The X/Open XNFS Specification
 * defined the error status in this case as NFSERR_NXIO.  This
 * appears to be inconsistent with existing server practice.)
 *
 * If the server does not support symbolic links, the status
 * NFS3ERR_NOTSUPP should be returned.
 *
 * On entry, the arguments in READLINK3args are as follows:
 *
 *	symlink		The file handle for the symbolic link
 *			object.
 *
 * On a successful return, the results in READLINK3resok
 * are as follows:
 *
 *	data		The data associated with the symbolic
 *			link object.
 *	symlink_attributes The attributes for the symbolic link object.
 *
 * On certain errors (as shown in the discriminated union
 * READLINK3res below) the following data is returned in the
 * structure READLINK3resfail:
 *
 *	symlink_attributes The attributes for the symbolic link object.
 *
 * On other errors, no results are returned.
 *
 * The following structure defines the arguments to the
 * procedure READLINK:
 */
struct READLINK3args {
	nfs_fh3		symlink;
};
/*
 * The following union defines the results from invoking the
 * procedure READLINK.
 */
struct READLINK3resok {
	post_op_attr	symlink_attributes;
	nfspath3	data;
};

struct READLINK3resfail {
	post_op_attr	symlink_attributes;
};

union READLINK3res switch (nfsstat3 status) {
case NFS3_OK:
	READLINK3resok	resok;
default:
	READLINK3resfail	resfail;
};

/*
 * NFS V3 procedure NFSPROC3_READ (index 6).
 *
 * The function of procedure READ is to read data from a
 * file.
 *
 * On entry, the arguments in READ3args are as follows:
 *
 *	file		The file handle for the file from
 *			which data is to be read.  This must
 *			identify a file system object of
 *			type NF3REG: if it does not, the
 *			status NFS3ERR_INVAL is returned.
 *			(Note: the NFS V2 server returns
 *			NFSERR_ISDIR, which is wrong.)
 *	offset		The position within the file at
 *			which the read is to begin.  If
 *			this is greater than or equal to
 *			the size of the file, the status
 *			NFS3_OK is returned with 'count' set
 *			to 0 and 'eof' set to TRUE.
 *	count		The number of bytes of data which
 *			are to be read.  The server may
 *			elect to return fewer bytes than
 *			requested.  If 'count' is 0, the
 *			READ will succeed and return 0
 *			bytes of data.
 *
 * On a successful return, the results in READ3resok
 * are as follows:
 *
 *	file_attributes	The attributes of the file after
 *			the completion of the read.
 *	count		The number of bytes of data
 *			returned by the read.  Note that this
 *			is strictly redundant, since the
 *			'data' is a counted data item.
 *	eof		If the read ended at the end-of-file
 *			(formally, if 'READ3args.offset' plus
 *			'READ3resok.count' is equal to the
 *			size of the file), 'eof' is returned
 *			as TRUE; otherwise it is FALSE.  Note
 *			that a READ of an empty file will
 *			always return 'eof' as TRUE.
 *	data		The data read from the file.
 *
 * On certain errors (as shown in the discriminated union
 * READ3res below) the following data is returned in the
 * structure READ3resfail:
 *
 *	file_attributes	The attributes of the file.
 *
 * On other errors, no results are returned.
 *
 * The following structure defines the arguments to the
 * procedure READ:
 */
struct READ3args {
	nfs_fh3		file;
	offset3		offset;
	count3		count;
};
/*
 * The following union defines the results from invoking the
 * procedure READ.
 */
struct READ3resok {
	post_op_attr	file_attributes;
	count3		count;
	bool		eof;
	opaque		data<>;
};

struct READ3resfail {
	post_op_attr	file_attributes;
};

union READ3res switch (nfsstat3 status) {
case NFS3_OK:
	READ3resok	resok;
default:
	READ3resfail	resfail;
};

/*
 * NFS V3 procedure NFSPROC3_WRITE (index 7).
 *
 * The function of procedure WRITE is to write data to
 * a file.
 *
 * On entry, the arguments in WRITE3args are as follows:
 *
 *	file		The file handle for the file to
 *			which data is to be written.  This
 *			must identify a file system object
 *			of type NF3REG: if it does not, the
 *			status NFS3ERR_INVAL is returned.
 *			(Note: the NFS V2 server returns
 *			NFSERR_ISDIR, which is wrong.)
 *	offset		The position within the file at
 *			which the write is to begin.
 *	count		The number of bytes of data which
 *			are to be written.  If 'count' is 0, the
 *			WRITE will succeed modulo the normal
 *			access checks.
 *	stable		If 'stable' is UNSTABLE, the server is
 *			is not forced to commit any part of the
 *			data or metadata of the file to stable
 *			storage before returning.  The only
 *			commitment made by the server is that it
 *			will not destroy any of the data before
 *			flushing it to stable storage without
 *			changing the value of the write verifier.
 *			If 'stable' is DATA_SYNC, the server
 *			should flush the data and any metadata
 *			required to retrieve the data to stable
 *			storage before returning.  Any metadata
 *			not required to retrieve the data may
 *			or may not be flushed to stable storage
 *			before returning but should not be
 *			destroyed without changing the value of
 *			the write verifier.
 *			If 'stable' is FILE_SYNC, the server
 *			should flush all data and all modified
 *			metadata associated with the file to
 *			stable storage before returning.  This
 *			corresponds to NFS V2 semantics.
 *			Note: the server is free to commit more
 *			data or metadata to stable storage before
 *			returning.  Given this, it is permissible
 *			for the server to flush data and metadata
 *			to stable storage even if 'stable' is
 *			UNSTABLE.  It is also permissible to
 *			simulate DATA_SYNC via FILE_SYNC.
 *			Note: it is *not* permissible to flush
 *			less data and metadata to stable storage
 *			before returning than was requested.
 *			This can easily result in the loss of data.
 *			This will be considered a protocol violation
 *			since such functionality should not be deemed
 *			required for this version of the protocol.
 *	data		The data to be written to the file.
 *
 * On a successful return, the results in WRITE3resok
 * are as follows:
 *
 *	file_wcc	Weak cache consistency data for the
 *			file.  This is described in detail
 *			in the comments preceding the
 *			definition of wcc_data.  For a client
 *			which requires only the post-write
 *			file attributes, these can be
 *			found in 'file_wcc.after'.
 *	count		The number of bytes of data written
 *			to the file.  The server may choose
 *			to write fewer butes than requested.
 *			In general, this is not recommended
 *			for performance reasons.
 *	committed	The status of the data and metadata
 *			should be returned via 'committed'.
 *			If the server flushed all of the data
 *			and metadata to stable storage, the
 *			value of 'committed' should be FILE_SYNC.
 *			If the server only flushed the data and
 *			enough metadata to retrieve the data to
 *			stable storage, the value of 'committed'
 *			should be DATA_SYNC.  Otherwise, the
 *			value of 'committed' should be set to
 *			UNSTABLE.
 *	verf		This is a cookie that the client can use to
 *			determine whether the server has rebooted
 *			between a call to WRITE and a subsequent
 *			call to COMMIT.  This cookie must be
 *			consistent during a single boot session and
 *			must be unique between boot sessions.  One
 *			suggestion would be to use the time that the
 *			server was last booted.
 *
 * On certain errors (as shown in the discriminated union
 * WRITE3res below) the following data is returned in the
 * structure WRITE3resfail:
 *
 *	file_wcc	Weak cache consistency data for the
 *			file.  This is described in detail
 *			in the comments preceding the
 *			definition of wcc_data.  For a client
 *			which requires only the post-WRITE
 *			file attributes, these can be
 *			found in 'file_wcc.after'.  Note that
 *			even though the WRITE failed, full
 *			wcc_data is returned to allow the
 *			client to determine whether the
 *			failing WRITE resulted in any change
 *			to the file.
 *
 * On other errors, no results are returned.
 *
 * The following structures defines the arguments to the
 * procedure WRITE:
 */
enum stable_how {
	UNSTABLE = 0,
	DATA_SYNC = 1,
	FILE_SYNC = 2
};

struct WRITE3args {
	nfs_fh3		file;
	offset3		offset;
	count3		count;
	stable_how	stable;
	opaque		data<>;
};
/*
 * The following union defines the results from invoking the
 * procedure WRITE.
 */
struct WRITE3resok {
	wcc_data	file_wcc;
	count3		count;
	stable_how	committed;
	writeverf3	verf;
};

struct WRITE3resfail {
	wcc_data	file_wcc;
};

union WRITE3res switch (nfsstat3 status) {
case NFS3_OK:
	WRITE3resok	resok;
default:
	WRITE3resfail	resfail;
};

/*
 * NFS V3 procedure NFSPROC3_CREATE (index 8).
 *
 * The function of procedure CREATE is to create a new regular
 * file.  Unlike NFS V2, in which certain fields in the initial
 * attributes structure were overloaded to indicate creation of
 * devices and fifos in addition to regular files, this procedure
 * only supports the creation of regular files.
 *
 * One aspect of the V3 CREATE procedure warrants particularly
 * careful consideration: the mechanism introduced to support
 * the reliable exclusive creation of regular files.  The
 * mechanism comes into play when 'how.mode' is EXCLUSIVE.
 * In this case, the client encodes in 'how.verf' a verifier
 * which can reasonably be expected to be unique.  [A combination
 * of client IP address and NFS RPC transaction identifier may
 * be appropriate.]
 *
 * If the file does not exist, the server creates the file
 * and stores the verifier on stable storage.  For file systems
 * which do not provide a mechanism for the storage of arbitrary
 * file attributes, the server may use one or more elements of
 * the file metadata to store the verifier.  (For this reason, an
 * exclusive file create may not include initial attributes.)
 * The server then returns with status NFS3_OK.
 *
 * If the file already exists, the server reconstructs the file's
 * verifier from the file metadata (or otherwise) and compares
 * it with the verifier in the request.  If they match, the server
 * treats the request as a success.  The request is presumed to
 * be a duplicate of an earlier, successful request for which
 * the reply was lost and which the server duplicate request
 * cache mechanism did not detect.
 *
 * If the verifiers do not match, the request is rejected with
 * the status NFS3ERR_EXIST.
 *
 * Once the client has performed a successful exclusive create, it
 * should issue a SETATTR to set the correct file attributes.  Until
 * it does so, it should not rely upon any of the file attributes.
 *
 * On entry, the arguments in CREATE3args are as follows:
 *
 *	where		Where the object is to be created:
 *	where.dir	The file handle for the directory in
 *			which the object is to be created.  If
 *			this does not identify a file system
 *			object of type NF3DIR, the status
 *			NFS3ERR_NOTDIR is returned.
 *	where.name	The name which is to be associated
 *			with the created object.  The general
 *			comments about file names apply.  The
 *			name may not have the value "." or ".."
 *			or the procedure will fail with the
 *			status NFS3ERR_EXIST.  Certain file systems
 *			may also disallow names which contain
 *			particular characters - for example, the
 *			UFS file system will reject a name which
 *			contains '/'.  The error status in this
 *			case is NFS3ERR_ACCES.
 *	how		A union defining how the server is to handle
 *			the file creation, together with any
 *			appropriate attributes:
 *	how.mode	One of UNCHECKED, GUARDED, and EXCLUSIVE.
 *			UNCHECKED means that the file should be created
 *			without checking for the existence of a duplicate
 *			file in the same directory.  In this case,
 *			'how.obj_attributes' is a sattr3 describing the
 *			initial attributes for the file.
 *
 *			GUARDED specifies that the server should check for
 *			the presence of a duplicate file before performing
 *			the create, and should fail the request with
 *			NFS3ERR_EXIST if a duplicate file exists.
 *			Otherwise the request is performed as
 *			described for UNCHECKED.
 *			[It must be emphasized that this does NOT
 *			provide exactly-once semantics.  In particular,
 *			if a reply is lost and the server's duplicate
 *			request cache does not detect the retransmission
 *			of the request, the procedure can fail with
 *			NFS3ERR_EXIST even though the create was
 *			performed successfully.]
 *
 *			EXCLUSIVE specifies that the server is to
 *			follow the exclusive creation semantics
 *			described above.  No attributes may be provided
 *			in this case, since the server may use the
 *			target file system attribute metadata to store
 *			the createverf3 'verf'.
 *
 * On a successful return, the results in CREATE3resok
 * are as follows:
 *
 *	obj		The file handle for the newly created 'where.name'.
 *	obj_attributes	The attributes for the object 'where.name'.
 *	dir_wcc		Weak cache consistency data for the directory
 *			'where.dir'.
 *
 * On certain errors (as shown in the discriminated union
 * CREATE3res below) the following data is returned in the
 * structure CREATE3resfail:
 *
 *	dir_wcc		Weak cache consistency data for the directory
 *			'where.dir'.  For a client which requires only
 *			the post-CREATE directory attributes, these can be
 *			found in 'dir_wcc.after'.  Note that even though
 *			the CREATE failed, full wcc_data is returned to
 *			allow the client to determine whether the
 *			failing CREATE resulted in any change
 *			to the directory.
 *
 * On other errors, no results are returned.
 *
 * The following structures define the arguments to the
 * procedure CREATE:
 */
/*
 * Create mode information
 */
enum createmode3 {
	UNCHECKED = 0,
	GUARDED = 1,
	EXCLUSIVE = 2
};

union createhow3 switch (createmode3 mode) {
case UNCHECKED:
case GUARDED:
	sattr3		obj_attributes;
case EXCLUSIVE:
	createverf3	verf;
};

struct CREATE3args {
	diropargs3	where;
	createhow3	how;
};
/*
 * The following union defines the results from invoking the
 * procedure CREATE.
 */
struct CREATE3resok {
	post_op_fh3	obj;
	post_op_attr	obj_attributes;
	wcc_data	dir_wcc;
};

struct CREATE3resfail {
	wcc_data	dir_wcc;
};

union CREATE3res switch (nfsstat3 status) {
case NFS3_OK:
	CREATE3resok	resok;
default:
	CREATE3resfail	resfail;
};

/*
 * NFS V3 procedure NFSPROC3_MKDIR (index 9).
 *
 * The function of procedure MKDIR is to create a new subdirectory
 * of a directory.  If there is an entry in the directory already
 * corresponding to the name of the subdirectory to be created,
 * the status NFS3ERR_EXISTS is returned.
 *
 * On entry, the arguments in MKDIR3args are as follows:
 *
 *	where		Where the object is to be created:
 *	where.dir	The file handle for the directory in
 *			which the subdirectory is to be created.
 *			If this does not identify a file system
 *			object of type NF3DIR, the status
 *			NFS3ERR_NOTDIR is returned.
 *	where.name	The name which is to be associated
 *			with the created subdirectory.  The general
 *			comments about file names apply.  The
 *			name may not have the value "." or ".."
 *			or the procedure will fail with the
 *			status NFS3ERR_EXIST.  Certain file systems
 *			may also disallow names which contain
 *			particular characters - for example, the
 *			UFS file system will reject a name which
 *			contains '/'.  The error status in this
 *			case is NFS3ERR_ACCES.
 *	dir_attributes	The initial attributes for the subdirectory.
 *
 * On a successful return, the results in MKDIR3resok
 * are as follows:
 *
 *	obj		The file handle for the newly created 'where.name'.
 *	obj_attributes	The attributes for the object 'where.name'.
 *	dir_wcc		Weak cache consistency data for the directory
 *			'where.dir'.
 *
 * On certain errors (as shown in the discriminated union
 * MKDIR3res below) the following data is returned in the
 * structure MKDIR3resfail:
 *
 *	dir_wcc		Weak cache consistency data for the directory
 *			'where.dir'.  For a client which requires only
 *			the post-MKDIR directory attributes, these can
 *			be found in 'dir_wcc.after'.  Note that even
 *			though the MKDIR failed, full wcc_data is
 *			returned to allow the client to determine
 *			whether the failing MKDIR resulted in any change
 *			to the directory.
 *
 * On other errors, no results are returned.
 *
 * The following structures define the arguments to the
 * procedure MKDIR:
 */
struct MKDIR3args {
	diropargs3	where;
	sattr3		attributes;
};
/*
 * The following union defines the results from invoking the
 * procedure MKDIR.
 */
struct MKDIR3resok {
	post_op_fh3	obj;
	post_op_attr	obj_attributes;
	wcc_data	dir_wcc;
};

struct MKDIR3resfail {
	wcc_data	dir_wcc;
};

union MKDIR3res switch (nfsstat3 status) {
case NFS3_OK:
	MKDIR3resok	resok;
default:
	MKDIR3resfail	resfail;
};

/*
 * NFS V3 procedure NFSPROC3_SYMLINK (index 10).
 *
 * The function of procedure SYMLINK is to create a new file system
 * object of type symbolic link.  The actual file system node and
 * its contents are expected to be created in one single atomic
 * operation.  If there is an entry in the directory already
 * corresponding to the name of symlink to be created, the status
 * NFS3ERR_EXIST is returned.
 *
 * If the server does not support symbolic links, the status
 * NFS3ERR_NOTSUPP should be returned.
 *
 * On entry, the arguments in SYMLINK3args are as follows:
 *
 *	where		Where the symlink is to be created:
 *	where.dir	The file handle for the directory in
 *			which the symlink is to be created.
 *			If this does not identify a file system
 *			object of type NF3DIR, the status
 *			NFS3ERR_NOTDIR is returned.
 *	where.name	The name which is to be associated
 *			with the created symlink.  The general
 *			comments about file names apply.  The
 *			name may not have the value "." or ".."
 *			or the procedure will fail with the
 *			status NFS3ERR_EXIST.  Certain file systems
 *			may also disallow names which contain
 *			particular characters - for example, the
 *			UFS file system will reject a name which
 *			contains '/'.  The error status in this
 *			case is NFS3ERR_ACCES.
 *	symlink		A struct symlinkdata3 with the following components:
 *	symlink.symlink_attributes
 *			The initial attributes for the symbolic link.
 *	symlink.symlink_data
 *			An nfspath3 string containing the symbolic link data.
 *
 * On a successful return, the results in SYMLINK3resok
 * are as follows:
 *
 *	obj		The file handle for the newly created 'where.name'.
 *	obj_attributes	The attributes for the object 'where.name'.
 *	dir_wcc		Weak cache consistency data for the directory
 *			'where.dir'.
 *
 * On certain errors (as shown in the discriminated union
 * SYMLINK3res below) the following data is returned in the
 * structure SYMLINK3resfail:
 *
 *	dir_wcc		Weak cache consistency data for the directory
 *			'where.dir'.  For a client which requires only
 *			the post-SYMLINK directory attributes, these can
 *			be found in 'dir_wcc.after'.  Note that even
 *			though the SYMLINK failed, full wcc_data is
 *			returned to allow the client to determine
 *			whether the failing SYMLINK resulted in any change
 *			to the directory.
 *
 * On other errors, no results are returned.
 *
 * The following structures define the arguments to the
 * procedure SYMLINK:
 */
struct symlinkdata3 {
	sattr3		symlink_attributes;
	nfspath3	symlink_data;
};

struct SYMLINK3args {
	diropargs3	where;
	symlinkdata3	symlink;
};
/*
 * The following union defines the results from invoking the
 * procedure SYMLINK.
 */
struct SYMLINK3resok {
	post_op_fh3	obj;
	post_op_attr	obj_attributes;
	wcc_data	dir_wcc;
};

struct SYMLINK3resfail {
	wcc_data	dir_wcc;
};

union SYMLINK3res switch (nfsstat3 status) {
case NFS3_OK:
	SYMLINK3resok	resok;
default:
	SYMLINK3resfail	resfail;
};

/*
 * NFS V3 procedure NFSPROC3_MKNOD (index 11).
 *
 * The function of procedure MKNOD is to create a new file system
 * object.  Unlike NFS V2, in which fields in the CREATE arguments
 * were overloaded to indicate the creation of certain types of
 * objects, the V3 protocol defines a new procedure for the
 * creation of `special' files.  If the server does not support
 * objects of type 'what.type' or if 'what.type' does not refer
 * to a legal type, the procedure will fail with the status
 * NFS3ERR_BADTYPE.
 *
 * If the server can not support any of NF3CHR, NF3BLK, NF3SOCK,
 * NF3FIFO, the status NFS3ERR_NOTSUPP should be returned.
 *
 * On entry, the arguments in MKNOD3args are as follows:
 *
 *	where		Where the object is to be created:
 *	where.dir	The file handle for the directory in
 *			which the object is to be created.  If
 *			this does not identify a file system
 *			object of type NF3DIR, the status
 *			NFS3ERR_NOTDIR is returned.
 *	where.name	The name which is to be associated
 *			with the created object.  The general
 *			comments about file names apply.  The
 *			name may not have the value "." or ".."
 *			or the procedure will fail with the
 *			status NFS3ERR_EXIST.  Certain file systems
 *			may also disallow names which contain
 *			particular characters - for example, the
 *			UFS file system will reject a name which
 *			contains '/'.  The error status in this
 *			case is NFS3ERR_ACCES.
 *	what		A union identifying the type of object to
 *			be created, together with the data
 *			and attributes appropriate to the type
 *			of object:
 *	what.type	The type of the object.
 *
 * When creating a character special file ('what.type' = NF3CHR)
 * or a block special file ('what.type' = NF3BLK), 'what' includes:
 *
 *	device		A struct devicedata3 with the following components:
 *	device.dev_attributes
 *			The initial attributes for the special file.
 *	device.spec	The major/minor pair in 'device.spec.specdata1'
 *			and 'device.spec.specdata2' respectively.
 *
 * When creating a Unix-mode socket ('what.type' = NF3SOCK)
 * or a FIFO ('what.type' = NF3FIFO), 'what' includes:
 *
 *	pipe_attributes	The initial attributes for the file.
 *
 * On a successful return, the results in MKNOD3resok
 * are as follows:
 *
 *	obj		The file handle for the newly created 'where.name'.
 *	obj_attributes	The attributes for the object 'where.name'.
 *	dir_wcc		Weak cache consistency data for the directory
 *			'where.dir'.
 *
 * On certain errors (as shown in the discriminated union
 * MKNOD3res below) the following data is returned in the
 * structure MKNOD3resfail:
 *
 *	dir_wcc		Weak cache consistency data for the directory
 *			'where.dir'.  For a client which requires only
 *			the post-MKNOD directory attributes, these can
 *			be found in 'dir_wcc.after'.  Note that even
 *			though the MKNOD failed, full wcc_data is
 *			returned to allow the client to determine
 *			whether the failing MKNOD resulted in any change
 *			to the directory.
 *
 * On other errors, no results are returned.
 *
 * The following structures define the arguments to the
 * procedure MKNOD:
 */
/*
 * Object creation per-type arguments
 */
struct devicedata3 {
	sattr3		dev_attributes;
	specdata3	spec;
};

union mknoddata3 switch (ftype3 type) {
case NF3CHR:
case NF3BLK:
	devicedata3	device;
case NF3SOCK:
case NF3FIFO:
	sattr3		pipe_attributes;
default:
	void;
};

struct MKNOD3args {
	diropargs3	where;
	mknoddata3	what;
};
/*
 * The following union defines the results from invoking the
 * procedure MKNOD.
 */
struct MKNOD3resok {
	post_op_fh3	obj;
	post_op_attr	obj_attributes;
	wcc_data	dir_wcc;
};

struct MKNOD3resfail {
	wcc_data	dir_wcc;
};

union MKNOD3res switch (nfsstat3 status) {
case NFS3_OK:
	MKNOD3resok	resok;
default:
	MKNOD3resfail	resfail;
};

/*
 * NFS V3 procedure NFSPROC3_REMOVE (index 12).
 *
 * The function of procedure REMOVE is to remove an entry from
 * a directory.  If the directory entry is the last reference to
 * the corresponding file system object, the object will be
 * removed.  The concept of "last reference" is server specific.
 * However, if the 'nlink' field in the previous attributes of the
 * object had the value 1, the client should not rely on being
 * able to refer to the object via a file handle.  Likewise, the
 * client should not rely on the resources (disk space, directory
 * entry, etc.) formerly associated with the object becoming
 * immediately available.
 *
 * On entry, the arguments in REMOVE3args are as follows:
 *
 *	object		A diropargs structure identifying the
 *			directory entry to be removed:
 *	object.dir	The file handle for the directory from
 *			which the entry is to be removed.  If
 *			this does not identify a file system
 *			object of type NF3DIR, the status
 *			NFS3ERR_NOTDIR is returned.
 *	object.name	The name of the entry to be removed.  The
 *			general comments about file names apply.
 *			If the name is "." the request will fail
 *			with NFS3ERR_INVAL.  If it is "..", the
 *			procedure will fail with NFS3ERR_EXIST.
 *			If the directory does not contain an
 *			entry with the (possibly truncated) name,
 *			the REMOVE fails with the status
 *			NFS3ERR_NOENT.
 *
 * On a successful return, the results in REMOVE3resok
 * are as follows:
 *
 *	dir_wcc		Weak cache consistency data for the
 *			directory 'object.dir'.
 *
 * On certain errors (as shown in the discriminated union
 * REMOVE3res below) the following data is returned in the
 * structure REMOVE3resfail:
 *
 *	dir_wcc		Weak cache consistency data for the
 *			directory 'object.dir'.  For a client
 *			which requires only the post-REMOVE
 *			directory attributes, these can be
 *			found in 'file_wcc.after'.  Note that
 *			even though the REMOVE failed, full
 *			wcc_data is returned to allow the
 *			client to determine whether the
 *			failing REMOVE resulted in any change
 *			to the directory.
 *
 * On other errors, no results are returned.
 *
 * The following structure defines the arguments to the
 * procedure REMOVE:
 */
struct REMOVE3args {
	diropargs3	object;
};
/*
 * The following union defines the results from invoking the
 * procedure REMOVE.
 */
struct REMOVE3resok {
	wcc_data	dir_wcc;
};

struct REMOVE3resfail {
	wcc_data	dir_wcc;
};

union REMOVE3res switch (nfsstat3 status) {
case NFS3_OK:
	REMOVE3resok	resok;
default:
	REMOVE3resfail	resfail;
};

/*
 * NFS V3 procedure NFSPROC3_RMDIR (index 13).
 *
 * The function of procedure RMDIR is to remove a subdirectory from
 * a directory.  If the directory entry is the last reference to
 * the corresponding file system object, the object will be
 * removed.  The concept of "last reference" is server specific.
 * However, if the 'nlink' field in the previous attributes of the
 * object had the value 1, the client should not rely on being
 * able to refer to the object via a file handle.  Likewise, the
 * client should not rely on the resources (disk space, directory
 * entry, etc.) formerly associated with the object becoming
 * immediately available.
 *
 * On entry, the arguments in RMDIR3args are as follows:
 *
 *	object		A diropargs structure identifying the
 *			subdirectory to be removed:
 *	object.dir	The file handle for the directory from
 *			which the subdirectory is to be removed.
 *			If this does not identify a file system
 *			object of type NF3DIR, the status
 *			NFS3ERR_NOTDIR is returned.
 *	object.name	The name of the subdirectory to be removed.
 *			The general comments about file names apply.
 *			If the name is "." the request will fail
 *			with NFS3ERR_INVAL.  If it is "..", the
 *			procedure will fail with NFS3ERR_EXIST.
 *			If the directory does not contain an
 *			entry with the (possibly truncated) name,
 *			the RMDIR fails with the status
 *			NFS3ERR_NOENT.
 *
 * On a successful return, the results in RMDIR3resok
 * are as follows:
 *
 *	dir_wcc		Weak cache consistency data for the
 *			directory 'object.dir'.
 *
 * On certain errors (as shown in the discriminated union
 * RMDIR3res below) the following data is returned in the
 * structure RMDIR3resfail:
 *
 *	dir_wcc		Weak cache consistency data for the
 *			directory 'object.dir'.  For a client
 *			which requires only the post-REMOVE
 *			directory attributes, these can be
 *			found in 'file_wcc.after'.  Note that
 *			even though the RMDIR failed, full
 *			wcc_data is returned to allow the
 *			client to determine whether the
 *			failing RMDIR resulted in any change
 *			to the directory.
 *
 * On other errors, no results are returned.
 *
 * The following structure defines the arguments to the
 * procedure RMDIR:
 */
struct RMDIR3args {
	diropargs3	object;
};
/*
 * The following union defines the results from invoking the
 * procedure RMDIR.
 */
struct RMDIR3resok {
	wcc_data	dir_wcc;
};

struct RMDIR3resfail {
	wcc_data	dir_wcc;
};

union RMDIR3res switch (nfsstat3 status) {
case NFS3_OK:
	RMDIR3resok	resok;
default:
	RMDIR3resfail	resfail;
};

/*
 * NFS V3 procedure NFSPROC3_RENAME (index 14).
 *
 * The function of procedure RENAME is to rename the file identified
 * by 'from.name' in the directory 'from.dir' to 'to.name' in the
 * directory 'to.dir'.  The operation is required to be atomic on
 * the server.  'to.dir' and 'from.dir' must reside on the same
 * filesystem on the server, or the operation will fail with
 * the status NFS3ERR_XDEV.  ("On the same filesystem" means that
 * the 'fsid' fields in the attributes for the directories are the
 * same.) Even though the operation is atomic, the status NFS3ERR_MLINK
 * may be returned if the server used a "link/unlink" sequence
 * internally.
 *
 * On entry, the arguments in RENAME3args are as follows:
 *
 *	from		A diropargs structure identifying the
 *			source (the filesystem object to be renamed):
 *	from.dir	The file handle for the directory from
 *			which the entry is to be removed.  If
 *			this does not identify a file system
 *			object of type NF3DIR, the status
 *			NFS3ERR_NOTDIR is returned.
 *	from.name	The name of the entry which identifies the
 *			object to be renamed.  The general comments
 *			about file names apply.  The name may not
 *			have the value "." or "..", or the procedure
 *			will fail with the status NFS3ERR_INVAL.
 *			If the directory does not contain an
 *			entry with the (possibly truncated) name,
 *			the RENAME fails with the status
 *			NFS3ERR_NOENT.  If the entry is an alias
 *			for 'from.dir', the RENAME fails with
 *			NFS3ERR_INVAL.
 *	to		A diropargs structure identifying the
 *			target (the new name for the object):
 *	to.dir		The file handle for the directory where
 *			the new entry is to be created.  If
 *			this does not identify a file system
 *			object of type NF3DIR, the status
 *			NFS3ERR_NOTDIR is returned.
 *	to.name		The new name for the object.  The general comments
 *			about file names apply.  The name may not
 *			have the value "." or "..", or the procedure
 *			will fail with the status NFS3ERR_INVAL
 *			Certain file systems may also disallow names
 *			which contain particular characters - for
 *			example, the UFS file system will reject a
 *			name which contains '/'.  In this case, the
 *			RENAME fails with the status NFS3ERR_ACCES.
 *
 *			If the directory 'to.dir' already contains an
 *			entry with the (possibly truncated) 'to.name',
 *			the source object must be compatible with the
 *			target: either both are non-directories, or
 *			both are directories and the target must be
 *			empty.  If compatible, the existing target is
 *			removed before the rename occurs.
 *
 * On a successful return, the results in RENAME3resok
 * are as follows:
 *
 *	fromdir_wcc	Weak cache consistency data for the directory
 *			'from.dir'.
 *	todir_wcc	Weak cache consistency data for the directory
 *			'to.dir'.
 *
 * On certain errors (as shown in the discriminated union
 * RENAME3res below) the following data is returned in the
 * structure RENAME3resfail:
 *
 *	fromdir_wcc	Weak cache consistency data for the directory
 *			'from.dir'.
 *	todir_wcc	Weak cache consistency data for the directory
 *			'to.dir'.
 *
 * On other errors, no results are returned.
 *
 * The following structure defines the arguments to the
 * procedure RENAME:
 */
struct RENAME3args {
	diropargs3	from;
	diropargs3	to;
};
/*
 * The following union defines the results from invoking the
 * procedure RENAME.
 */
struct RENAME3resok {
	wcc_data	fromdir_wcc;
	wcc_data	todir_wcc;
};

struct RENAME3resfail {
	wcc_data	fromdir_wcc;
	wcc_data	todir_wcc;
};

union RENAME3res switch (nfsstat3 status) {
case NFS3_OK:
	RENAME3resok	resok;
default:
	RENAME3resfail	resfail;
};

/*
 * NFS V3 procedure NFSPROC3_LINK (index 15).
 *
 * The function of procedure LINK is to create a link to an
 * existing file system object.  'file' and 'link.dir' must reside on
 * the same filesystem on the server, or the operation will fail with
 * the status NFS3ERR_XDEV.  ("On the same filesystem" means that
 * the 'fsid' fields in the attributes for 'file' and 'link.dir' are the
 * same.)
 *
 * On entry, the arguments in LINK3args are as follows:
 *
 *	file		The file handle for the existing file
 *			system object.
 *	link		Where the link is to be created:
 *	link.dir	The file handle for the directory in
 *			which the link is to be created.  If
 *			this does not identify a file system
 *			object of type NF3DIR, the status
 *			NFS3ERR_NOTDIR is returned.
 *	link.name	The name which is to be associated
 *			with the created link.  The general comments
 *			about file names apply.  The name may not
 *			have the value "." or "..", or the procedure
 *			will fail with the status NFS3ERR_INVAL
 *			Certain file systems may also disallow names
 *			which contain particular characters - for
 *			example, the UFS file system will reject a
 *			name which contains '/'.  In this case, the
 *			LINK fails with the status NFS3ERR_ACCES.
 *
 * On a successful return, the results in LINK3resok
 * are as follows:
 *
 *	file_attributes	The attributes of the filesystem object
 *			identified by 'file'.
 *	linkdir_wcc	Weak cache consistency data for the directory
 *			'link.dir'.
 *
 * On certain errors (as shown in the discriminated union
 * LINK3res below) the following data is returned in the
 * structure LINK3resfail:
 *
 *	file_attributes	The attributes of the filesystem object
 *			identified by 'file'.
 *	linkdir_wcc	Weak cache consistency data for the directory
 *			'link.dir'.
 *
 * On other errors, no results are returned.
 *
 * The following structure defines the arguments to the
 * procedure LINK:
 */
struct LINK3args {
	nfs_fh3		file;
	diropargs3	link;
};
/*
 * The following union defines the results from invoking the
 * procedure LINK.
 */
struct LINK3resok {
	post_op_attr	file_attributes;
	wcc_data	linkdir_wcc;
};

struct LINK3resfail {
	post_op_attr	file_attributes;
	wcc_data	linkdir_wcc;
};

union LINK3res switch (nfsstat3 status) {
case NFS3_OK:
	LINK3resok	resok;
default:
	LINK3resfail	resfail;
};

/*
 * NFS V3 procedure NFSPROC3_READDIR (index 16).
 *
 * The function of procedure READDIR is to retrieve a sequence
 * of entries from a file system directory and return the name and
 * fileid for each, together with information to allow the client
 * to request additional directory entries in a subsequent READDIR.
 *
 * In NFS V2, each directory entry returned included a 'cookie'
 * identifying a point in the directory.  By including this
 * cookie in a subsequent READDIR, the client could resume the
 * directory search at any point in the directory.
 *
 * One problem with this scheme was that there was no easy
 * way for a server to verify that a cookie was valid.  If
 * two READDIRs were separated by one or more operations which
 * changed the directory in some way (for example, reordering
 * or compressing it), it was possible that the second READDIR
 * could miss entries, or process entries more than once.  If the
 * cookie was no longer usable (e.g. pointing into the middle of
 * a directory entry), the server would have to indicate that the
 * search had reached the end of the directory, even though more
 * entries remained, and there was no way that the client could
 * distinguish this from a valid end-of-directory.
 *
 * In NFS V3, each READDIR request includes both a cookie and a
 * cookie verifier.  For the first call, both are set to 0.
 * The response includes a new cookie verifier, together with
 * a cookie per entry.  For subsequent READDIRs, the client must
 * present both the cookie and the corresponding cookie verifier.
 * If the server detects that the cookie is no longer valid, it
 * will reject the request with the status NFS3ERR_BAD_COOKIE.
 *
 * [One implementation of the cookie verifier mechanism might be
 * for the server to use the modification time of the directory.
 * This might be overly restrictive, however.  A better approach
 * would be to record the time of the last direction modification
 * which changed the directory organization in a way which would
 * make it impossible to reliably interpret a cookie.]
 *
 * On entry, the arguments in READDIR3args are as follows:
 *
 *	dir		The file handle for the directory to be searched.
 *			If this file handle does not refer to a directory,
 *			the error, NFS3ERR_NOTDIR should be returned.
 *	cookie		On the first request to search a directory, this
 *			should be 0.  On subsequent requests, it should
 *			be a cookie returned by the server.
 *	cookieverf	On the first request to search a directory, this
 *			should be 0.  On subsequent requests, it should
 *			be a cookieverf returned by the server.  The
 *			cookieverf must match that returned by the
 *			READDIR in which the 'cookie' was required.
 *	count		The maximum size of the READDIR3resok structure,
 *			in bytes.  The size must include all XDR overhead.
 *			The server is free to return less than 'count'
 *			bytes of data.  If 'count' is too small to
 *			return a single directory entry, the request
 *			will fail with the status NFS3ERR_TOOSMALL.
 *
 * On a successful return, the results in READDIR3resok
 * are as follows:
 *
 *	dir_attributes	The attributes of the directory 'dir'.
 *	cookieverf	The cookie verifier which should accompany
 *			any cookie from 'reply'.
 *	reply		The directory list:
 *	reply.entries	Zero or more directory entries.
 *	reply.eof	TRUE if the last member of 'reply.entries'
 *			is the last entry in the directory OR the
 *			list 'reply.entries' is empty and the 'cookie'
 *			corresponded to the end of the directory.
 *			If FALSE, there may be more entries to read.
 *
 * On certain errors (as shown in the discriminated union
 * READDIR3res below) the following data is returned in the
 * structure READDIR3resfail:
 *
 *	dir_attributes	The attributes of the directory 'dir'.
 *
 * On other errors, no results are returned.
 *
 * The following structure defines the arguments to the
 * procedure READDIR:
 */
struct READDIR3args {
	nfs_fh3		dir;
	cookie3		cookie;
	cookieverf3	cookieverf;
	count3		count;
};
/*
 * The following structures define the results of invoking the
 * procedure READDIR.
 */
struct entry3 {
	fileid3		fileid;
	filename3	name;
	cookie3		cookie;
	entry3		*nextentry;
};

struct dirlist3 {
	entry3		*entries;
	bool		eof;
};

struct READDIR3resok {
	post_op_attr	dir_attributes;
	cookieverf3	cookieverf;
	dirlist3	reply;
};

struct READDIR3resfail {
	post_op_attr	dir_attributes;
};

union READDIR3res switch (nfsstat3 status) {
case NFS3_OK:
	READDIR3resok	resok;
default:
	READDIR3resfail	resfail;
};

/*
 * NFS V3 procedure NFSPROC3_READDIRPLUS (index 17).
 *
 * The function of procedure READDIRPLUS is to retrieve a sequence
 * of entries from a file system directory and return complete
 * information about each, together with information to allow the client
 * to request additional directory entries in a subsequent READDIRPLUS.
 *
 * READDIRPLUS differs from READDIR only in the amount of information
 * returned for each entry.  In READDIR, each entry returns a file
 * name, a fileid, and a cookie for the next entry.  In READDIRPLUS,
 * each entry returns the same, plus the attributes and file handle.
 *
 * On entry, the arguments in READDIRPLUS3args are as follows:
 *
 *	dir		The file handle for the directory to be searched.
 *			If this file handle does not refer to a directory,
 *			the error, NFS3ERR_NOTDIR should be returned.
 *	cookie		On the first request to search a directory, this
 *			should be 0.  On subsequent requests, it should
 *			be a cookie returned by the server.
 *	cookieverf	On the first request to search a directory, this
 *			should be 0.  On subsequent requests, it should
 *			be a cookieverf returned by the server.  The
 *			cookieverf must match that returned by the
 *			READDIRPLUS in which the 'cookie' was required.
 *	dircount	The maximum number of bytes of directory information
 *			returned.  This number should not include the
 *			attributes and file handles portions of the results.
 *	maxcount	The maximum size of the READDIRPLUS3resok structure,
 *			in bytes.  The size must include all XDR overhead.
 *			The server is free to return less than 'count'
 *			bytes of data.  If 'count' is too small to
 *			return a single directory entry, the request
 *			will fail with the status NFS3ERR_TOOSMALL.
 *
 * On a successful return, the results in READDIRPLUS3resok
 * are as follows:
 *
 *	dir_attributes	The attributes of the directory 'dir'.
 *	cookieverf	The cookie verifier which should accompany
 *			any cookie from 'reply'.
 *	reply		The directory list:
 *	reply.entries	Zero or more directory entries.
 *	reply.eof	TRUE if the last member of 'reply.entries'
 *			is the last entry in the directory OR the
 *			list 'reply.entries' is empty and the 'cookie'
 *			corresponded to the end of the directory.
 *			If FALSE, there may be more entries to read.
 *
 * On errors, the following data is returned in READDIRPLUS3resfail:
 *
 *	dir_attributes	The attributes of the directory 'dir'.
 */

/*
 * The following structure defines the arguments to the
 * procedure READDIRPLUS:
 */
struct READDIRPLUS3args {
	nfs_fh3		dir;
	cookie3		cookie;
	cookieverf3	cookieverf;
	count3		dircount;
	count3		maxcount;
};

/*
 * The following structures define the results of invoking the
 * procedure READDIRPLUS.
 */
struct entryplus3 {
	fileid3		fileid;
	filename3	name;
	cookie3		cookie;
	post_op_attr	name_attributes;
	post_op_fh3	name_handle;
	entryplus3	*nextentry;
};

struct dirlistplus3 {
	entryplus3	*entries;
	bool		eof;
};

struct READDIRPLUS3resok {
	post_op_attr	dir_attributes;
	cookieverf3	cookieverf;
	dirlistplus3	reply;
};

struct READDIRPLUS3resfail {
	post_op_attr	dir_attributes;
};

union READDIRPLUS3res switch (nfsstat3 status) {
case NFS3_OK:
	READDIRPLUS3resok resok;
default:
	READDIRPLUS3resfail resfail;
};

/*
 * NFS V3 procedure NFSPROC3_FSSTAT (index 18).
 *
 * The function of procedure FSSTAT is to retrieve volatile
 * file system state information.
 *
 * On entry, the arguments in FSSTAT3args are as follows:
 *
 *	fsroot		The file handle identifying the mount
 *			point for the file system, as provided
 *			by the mount service on the server.
 *
 * On a successful return, the results in FSSTAT3resok
 * are as follows:
 *
 *	obj_attributes	The attributes of the mount point file system
 *			object.
 *	tbytes		The total size of the file system, in bytes.
 *	fbytes		The amount of free space in the file system.
 *	abytes		The amount of free space which is available
 *			to the user corresponding to the authentication
 *			information in the RPC.  (This reflects space
 *			which is reserved by the file system; it does
 *			not reflect any quota system implemented by
 *			the server.)
 *	tfiles		The total number of file slots in the file
 *			system.  (On a Unix server, this corresponds to
 *			the number of inodes configured.)
 *	ffiles		The number of free file slots in the file
 *			system.
 *	afiles		The number of free file slots which are available
 *			to the user corresponding to the authentication
 *			information in the RPC.  (This reflects slots
 *			which are reserved by the file system; it does
 *			not reflect any quota system implemented by
 *			the server.)
 *	invarsec	A measure of file system volatility: the number of
 *			seconds for which the file system is not expected
 *			to change.  For a volatile, frequently updated
 *			file system this will be 0.  For an immutable
 *			file system, such as a CD-ROM, this might be
 *			MAXUNSIGNEDINT.  For file systems which are
 *			infrequently modified (e.g. one containing
 *			local executables and on-line documentation)
 *			a value corresponding to a few hours or days
 *			might be used.  The client may use this as a hint
 *			in tuning its cache management.
 *
 * On any error, no results are returned.
 *
 * The following structure defines the arguments to the
 * procedure FSSTAT:
 */
struct FSSTAT3args {
	nfs_fh3		fsroot;
};
/*
 * The following union defines the results from invoking the
 * procedure FSSTAT.
 */
struct FSSTAT3resok {
	post_op_attr	obj_attributes;
/* total/free space information */
	size3		tbytes;	/* total bytes in file system */
	size3		fbytes;	/* free bytes in file system */
	size3		abytes;	/* bytes available to this user */
	size3		tfiles;	/* total file slots in file system */
	size3		ffiles;	/* free file slots in file system */
	size3		afiles;	/* file slots available to this user */
/* volatility */
	uint32		invarsec;/* estimated seconds FS invariance */
};

struct FSSTAT3resfail {
	post_op_attr	obj_attributes;
};

union FSSTAT3res switch (nfsstat3 status) {
case NFS3_OK:
	FSSTAT3resok	resok;
default:
	FSSTAT3resfail	resfail;
};

/*
 * NFS V3 procedure NFSPROC3_FSINFO (index 19).
 *
 * The function of procedure FSINFO is to retrieve non-volatile
 * information about the file system and indirectly about the
 * NFS server implementation.
 *
 * On entry, the arguments in FSINFO3args are as follows:
 *
 *	fsroot		The file handle identifying the mount
 *			point for the file system, as provided
 *			by the mount service on the server.
 *
 * On a successful return, the results in FSINFO3resok
 * are as follows:
 *
 *	obj_attributes	The attributes of the mount point file system
 *			object.
 *	rtmax		The maximum size of READ request supported
 *			by the server.  Any READ with a 'count' greater
 *			than 'rtmax' will result in a short read of
 *			'rtmax' bytes or less.
 *	rtpref		The preferred size of READ request.  This
 *			should be the same as 'rtmax' unless there is
 *			a clear benefit in performance or efficiency.
 *	rtmult		The suggested multiple for the size of a
 *			READ request.
 *	wtmax		The maximum size of WRITE request supported
 *			by the server.  In general, the client
 *			must respect 'wtmax' since there is no guarantee
 *			that a server can cope with a larger WRITE.
 *	wtpref		The preferred size of WRITE request.  This
 *			should be the same as 'wtmax' unless there is
 *			a clear benefit in performance or efficiency.
 *	wtmult		The suggested multiple for the size of a
 *			WRITE request.
 *	dtpref		The preferred size of READDIR or READDIRPLUS
 *			request.
 *	maxfilesize	The maximum size of a file on the file system.
 *	time_delta	The server time granularity.  When setting
 *			a file time using SETATTR, the server guarantees
 *			only to preserve times to this accuracy.
 *			If this is {0,1}, the server can support
 *			nanosecond times.  {0,1000000} denotes millisecond
 *			precision, while {1,0} indicates that times
 *			are accurate only to the nearest second.
 *	properties	A bitmask of file system properties.  The following
 *			values are presently defined:
 *	FSF3_LINK	If this bit is 1 (TRUE), the file system supports
 *			hard links.
 *	FSF3_SYMLINK	If this bit is 1 (TRUE), the file system supports
 *			symbolic links.
 *	FSF3_HOMOGENEOUS If this bit is 1 (TRUE), the pathconf information
 *			is identical for every file and directory in the
 *			filesystem.  If it is 0 (FALSE), the client
 *			should retrieve pathconf information for each
 *			file and directory as required.
 *	FSF3_CANSETTIME	If this bit is 1 (TRUE), the server will set the
 *			times for a file via SETATTR if requested (to the
 *			accuracy indicated by 'time_delta').  If it is 0
 *			(FALSE), the server cannot set times as requested.
 *
 * On any error, no results are returned.
 *
 * The following structure defines the arguments to the
 * procedure FSINFO:
 */
struct FSINFO3args {
	nfs_fh3		fsroot;
};
/*
 * The following union defines the results from invoking the
 * procedure FSINFO.
 */
struct FSINFO3resok {
	post_op_attr	obj_attributes;
/* transfer information */
	uint32		rtmax;	/* maximum read transfer size */
	uint32		rtpref;	/* preferred read transfer size */
	uint32		rtmult;	/* suggested multiple for read transfer */
	uint32		wtmax;	/* maximum write transfer size */
	uint32		wtpref;	/* preferred write transfer size */
	uint32		wtmult;	/* suggested multiple for write transfer */
	uint32		dtpref;	/* preferred directory read size */
/* file system limits */
	size3		maxfilesize;	/* maximum size of a file */
	nfstime3	time_delta;	/* server minimum time discrimination */
/* file system properties */
	uint32		properties;	/* bit mask of properties */
};

struct FSINFO3resfail {
	post_op_attr	obj_attributes;
};

/*
 * permanent properties bits
 *
 * Bit is ON (set) if facility is supported or property is true, or server
 * suggests that client treat file system as described
 */
const FSF3_LINK = 0x1;		/* (hard) links supported */
const FSF3_SYMLINK = 0x2;	/* symbolic links supported */
const FSF3_HOMOGENEOUS = 0x8;	/* pathconf information cannot vary per file */
const FSF3_CANSETTIME = 0x10;	/* server always sets file times as requested */

union FSINFO3res switch (nfsstat3 status) {
case NFS3_OK:
	FSINFO3resok	resok;
default:
	FSINFO3resfail	resfail;
};

/*
 * NFS V3 procedure NFSPROC3_PATHCONF (index 20).
 *
 * The function of procedure PATHCONF is to retrieve the "pathconf"
 * information for a file or directory.  If the FSF3_HOMOGENEOUS bit is
 * set in 'FSFINFOresok.properties', the pathconf information will
 * be the same for all files and directories.
 *
 * On entry, the arguments in PATHCONF3args are as follows:
 *
 *	object		The file handle for the object in question.
 *			To obtain the default pathconf information for
 *			the filesystem, 'object' should be set to the
 *			mount point file handle as returned by the
 *			mount service on the server.
 *
 * On a successful return, the results in PATHCONF3resok
 * are as follows:
 *
 *	obj_attributes	The attributes of 'object'.
 *	linkmax		The maximum number of links supported.
 *	name_max	The maximum length of a file name.
 *	no_trunc	If TRUE, the server will reject any NFS
 *			request which includes a name longer than
 *			'name_max' with NFS3ERR_NAMETOOLONG.  If FALSE,
 *			any overlength name will be silently truncated
 *			to 'name_max' bytes.
 *	chown_resticted	if TRUE, the server will reject any NFS
 *			request to change either the owner or the
 *			group associated with a file if the caller
 *			is not the super user.  (Uid 0.)
 *	case_insensitive
 *			If TRUE, the server does not distinguish
 *			case when interpreting filenames.
 *	case_preserving	If TRUE, the server will preserve the case
 *			of a name during a CREATE, RENAME or LINK.
 *
 * On any error, no results are returned.
 *
 * The following structure defines the arguments to the
 * procedure PATHCONF:
 */
struct PATHCONF3args {
	nfs_fh3		object;
};
/*
 * The following union defines the results from invoking the
 * procedure PATHCONF.
 */
struct PATHCONF3resok {
	post_op_attr	obj_attributes;
/* posix-style pathconf info */
	uint32		link_max;
	uint32		name_max;
	bool		no_trunc;
	bool		chown_restricted;
/* additional info about file names */
	bool		case_insensitive;
	bool		case_preserving;
};

struct PATHCONF3resfail {
	post_op_attr	obj_attributes;
};

union PATHCONF3res switch (nfsstat3 status) {
case NFS3_OK:
	PATHCONF3resok	resok;
default:
	PATHCONF3resfail	resfail;
};

/*
 * NFS V3 procedure NFSPROC3_COMMIT (index 21).
 *
 * The function of procedure COMMIT is to commit cached data on
 * the server to stable storage.
 *
 * On entry, the arguments in COMMIT3args are as follows:
 *
 *	file		The file handle for the file to which data is
 *			to be written.  This must identify a file
 *			system object of type NF3REG: if it does not,
 *			the status NFS3ERR_INVAL is returned.
 *	offset		The position within the file at which the flush
 *			is to begin.
 *	count		The number of bytes of data which are to be
 *			written.  If 'count' is 0, a flush from offset
 *			'offset' to the end of file is done.
 *
 * On a successful return, the results in COMMIT3resok are as
 * follows:
 *
 *	file_wcc	Weak cache consistency data for the file.  This
 *			is described in detail in the comments
 *			preceding the definition of wcc_data.  For a
 *			client which requires only the post-write file
 *			attributes, these can be found in
 *			'file_wcc.after'.
 *	verf		This is a cookie that the client can use to
 *			determine whether the server has rebooted
 *			between a call to WRITE and a subsequent
 *			call to COMMIT.  This number must be
 *			consistent during a single boot session and
 *			must be unique between boot sessions.  One
 *			suggestion would be to use the time that the
 *			server was last booted.
 *
 * On certain errors, the following data is returned in the structure
 * COMMIT3resfail:
 *
 *	file_wcc	Weak cache consistency data for the file.  This
 *			is described in detail in the comments
 *			preceding the definition of wcc_data.  For a
 *			client which requires only the post-write file
 *			attributes, these can be found in
 *			'file_wcc.after'.  Note that even though the
 *			COMMIT failed, full wcc_data is returned
 *			to allow the client to determine whether the
 *			file changed on the server between calls to
 *			WRITE and COMMIT.
 *
 * On other errors, no results are returned.
 *
 * The following structure defines the arguments to the procedure
 * COMMIT:
 */
struct COMMIT3args {
	nfs_fh3		file;
	offset3		offset;
	count3		count;
};
/*
 * The following union defines the results from invoking the
 * procedure COMMIT.
 */
struct COMMIT3resok {
	wcc_data	file_wcc;
	writeverf3	verf;
};

struct COMMIT3resfail {
	wcc_data	file_wcc;
};

union COMMIT3res switch (nfsstat3 status) {
case NFS3_OK:
	COMMIT3resok	resok;
default:
	COMMIT3resfail	resfail;
};

program NFS3_PROGRAM {
	version	NFS_V3 {
		void
		 NFSPROC3_NULL(void)			=	0;
		GETATTR3res
		 NFSPROC3_GETATTR(GETATTR3args)		=	1;
		SETATTR3res
		 NFSPROC3_SETATTR(SETATTR3args)		=	2;
		LOOKUP3res
		 NFSPROC3_LOOKUP(LOOKUP3args)		=	3;
		ACCESS3res
		 NFSPROC3_ACCESS(ACCESS3args)		=	4;
		READLINK3res
		 NFSPROC3_READLINK(READLINK3args)	=	5;
		READ3res
		 NFSPROC3_READ(READ3args)		=	6;
		WRITE3res
		 NFSPROC3_WRITE(WRITE3args)		=	7;
		CREATE3res
		 NFSPROC3_CREATE(CREATE3args)		=	8;
		MKDIR3res
		 NFSPROC3_MKDIR(MKDIR3args)		=	9;
		SYMLINK3res
		 NFSPROC3_SYMLINK(SYMLINK3args)		=	10;
		MKNOD3res
		 NFSPROC3_MKNOD(MKNOD3args)		=	11;
		REMOVE3res
		 NFSPROC3_REMOVE(REMOVE3args)		=	12;
		RMDIR3res
		 NFSPROC3_RMDIR(RMDIR3args)		=	13;
		RENAME3res
		 NFSPROC3_RENAME(RENAME3args)		=	14;
		LINK3res
		 NFSPROC3_LINK(LINK3args)		=	15;
		READDIR3res
		 NFSPROC3_READDIR(READDIR3args)		=	16;
		READDIRPLUS3res
		 NFSPROC3_READDIRPLUS(READDIRPLUS3args)	=	17;
		FSSTAT3res
		 NFSPROC3_FSSTAT(FSSTAT3args)		=	18;
		FSINFO3res
		 NFSPROC3_FSINFO(FSINFO3args)		=	19;
		PATHCONF3res
		 NFSPROC3_PATHCONF(PATHCONF3args)	=	20;
		COMMIT3res
		 NFSPROC3_COMMIT(COMMIT3args)		=	21;
	} = 3;
} = 100003;
